# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------
# Author: Philippe LABORIE, IBM Analytics, France Lab, Gentilly

"""
This module contains the objects representing CP Optimizer function expressions,
Stepwise and Piecewise Linear functions:

In particular, it defines the following classes:

 * :class:`CpoFunction`: the root class of all function expressions,
 * :class:`CpoSegmentedFunction`: for functions represented as a list of segments,
 * :class:`CpoStepFunction`: for functions represented as a list of steps.


Detailed description
--------------------
"""

from docplex.cp.catalog import Type_SegmentedFunction, Type_StepFunction
from docplex.cp.expression import CpoExpr, INTERVAL_MIN, INTERVAL_MAX
from docplex.cp.utils import *
import bisect


class CpoFunction(CpoExpr):
    """ Root class for step and segmented functions.
    """
    __slots__ = ('_s0',  # Initial slope of the function
                 '_v0',  # Initial value of the function
                 '_x',   # Abscissas
                 '_v',   # Values
                 '_s',   # Slopes
                 )

    def __init__(self, typ, s0=None, v0=None, x=None, v=None, s=None, name=None):
        """
        Args:
            typ: Function type.
            s0:  Initial slope of the function.
            v0:  Initial value of the function.
            x:   Abscissas (assumed to be sorted and all different).
            v:   Values.
            s:   Slopes.
        """
        super(CpoFunction, self).__init__(typ, name)
        if v0 is None:
            self._v0 = 0
        else:
            self._v0 = v0
        if s0 is None:
            self._s0 = 0
        else:
            # Test type is segmented function when some slope is specified
            assert (s0 == 0) or (typ == Type_SegmentedFunction), "Slopes are only supported for segmented functions"
            # Test there exists at least one abscissa
            assert s0 == 0 or (x is not None and 0 < len(x)), "No abscissa for initial slope"
            self._s0 = s0
        if x is None:
            self._x = []
        else:
            # Test abscissa integrality
            assert all(is_int(elt) for elt in x), "Abscissa value is not an integer"
            # Test abscissa list is strictly increasing
            assert all(x[i] < x[i + 1] for i in range(len(x) - 1)), "Abscissa values are not strictly increasing"
            self._x = x
        l = len(self._x)
        if v is None:
            self._v = []
        else:
            # Test value integrality in case of step function
            if typ == Type_StepFunction:
                assert all(is_int(elt) for elt in v), "Value is not an integer"
            self._v = v
        assert len(self._v) == l, "Lists x and v must have the same length"
        if s is None:
            self._s = [0] * len(self._x)
        else:
            # Test type is segmented function when some slope is specified
            assert all(
                v == 0 for v in s) or typ == Type_SegmentedFunction, "Slopes are only supported for segmented functions"
            self._s = s
        assert len(self._s) == l, "Lists x and s must have the same length"

    def _get_expr_string(self):
        """ Get the string representing this expression, name excluded. """
        return "Function(...)"

    @property
    def v0(self):
        return self._v0

    @property
    def s0(self):
        return self._s0

    @property
    def x(self):
        return self._x

    @property
    def v(self):
        return self._v

    @property
    def s(self):
        return self._s

    @property
    def is_step_function(self):
        return self.type == Type_StepFunction

    @property
    def is_segmented_function(self):
        return self.type == Type_SegmentedFunction

    def get_value(self, t):
        """ Gets the value of the function.

        Gets the value of the function for an abscissa `t`.
        Complexity is in `O(log n)` for a function with `n` segments.

        Args:
            t (int or float): Abscissa value.

        Returns:
            Value of the function for abscissa `t`.
        """
        l = len(self._x)
        if l == 0:
            # Constant step function
            return self._v0
        if t < self._x[0]:
            # Abscissa t is on initial step
            return self._v0 - (self._x[0] - t) * self._s0
        # Abscissa is on another step
        i = bisect.bisect_right(self._x, t)
        if self._s is None:
            return self._v[i - 1]
        else:
            return self._v[i - 1] + (t - self._x[i - 1]) * self._s[i - 1]

    def copy(self, type=None):
        """ Creates and returns a copy of the function. """
        xcopy = self._x[:]
        vcopy = self._v[:]
        scopy = self._s[:]
        if type is None:
            tp = self.type
        else:
            tp = type
        return CpoFunction(tp, self._s0, self._v0, xcopy, vcopy, scopy)

    def set_slope(self, x1, x2, v1, s):
        """ Sets the value of the invoking function over an interval.

        Sets the value of the invoking function on interval `[x1,x2)` to be equal
        to equal to `f(x) = v1 + s * (x-x1)`.

        Args:
            x1 (int): Start of the interval.
            x2 (int): End of the interval.
            v1 (int or float): Function value at `x1`.
            s (int or float): Function slope on interval `[x1,x2)`.
        """
        assert s == 0 or not self.is_step_function, "Function not allowed on CpoStepFunction"
        assert is_int(x1) and is_int(x2), "Interval [x1,x2) must be integer"
        if x2 <= x1:
            return
        l = len(self._x)
        if l == 0:
            # Constant step function of value self._v0
            if not (s == 0 and v1 == self._v0):
                self._x.append(x1)
                self._v.append(v1)
                self._s.append(s)
                if x2 < INTERVAL_MAX:
                    self._x.append(x2)
                    self._v.append(self._v0)
                    self._s.append(0)
            return
        if x1 < self._x[0]:
            if self._s0 == s and self._v0 + s * (x1 - self._x[0]) == v1:
                # No change on initial segment
                if x2 <= self._x[0]:
                    # Completely inside initial segment
                    return
                # Make initial segment longer
                self._v0 += s * (x2 - self._x[0])
                i2 = bisect.bisect_right(self._x, x2) - 1
                if 0 < i2:
                    # Remove segments 0..i2-1 and shift arrays
                    del self._x[:i2]
                    del self._v[:i2]
                    del self._s[:i2]
                # Adjust beginning of first segment
                self._v[0] += (x2 - self._x[0]) * self._s[0]
                self._x[0] = x2
                return
            else:
                # Adjust value of initial segment
                v0 = self._v0
                x0 = self._x[0]
                self._v0 += self._s0 * (x1 - self._x[0])
                # Add new segment
                self._x.insert(0, x1)
                self._v.insert(0, v1)
                self._s.insert(0, s)
                if x2 < x0:
                    self._x.insert(1, x2)
                    self._v.insert(1, v0 + (x2 - x0) * self._s0)
                    self._s.insert(1, self._s0)
                    return
                elif x0 <= x2:
                    self.set_slope(x0, x2, v1 + (x0 - x1) * s, s)
                    return
        else:
            # x1 >= self._x[0]:
            i1 = bisect.bisect_right(self._x, x1) - 1
            i2 = bisect.bisect_right(self._x, x2) - 1
            if self._s[i1] == s and self._v[i1] + s * (x1 - self._x[i1]) == v1:
                # Same as current segment
                if i1 == i2:
                    return
                else:
                    self.set_slope(self._x[i1 + 1], x2, v1 + (self._x[i1 + 1] - x1) * s, s)
                    return
            else:
                xcurr = self._x[i1]
                vcurr = self._v[i1]
                scurr = self._s[i1]
                if x1 > self._x[i1]:
                    self._x.insert(i1 + 1, x1)
                    self._v.insert(i1 + 1, v1)
                    self._s.insert(i1 + 1, s)
                    i1 += 1
                    i2 += 1
                else:
                    assert x1 == self._x[i1]
                    self._x[i1] = x1
                    self._v[i1] = v1
                    self._s[i1] = s
                if i1 == i2:
                    self._x.insert(i1 + 1, x2)
                    self._v.insert(i1 + 1, vcurr + (x2 - xcurr) * scurr)
                    self._s.insert(i1 + 1, scurr)
                    return
                else:
                    # Adjust beginning of first segment
                    self._v[i2] += (x2 - self._x[i2]) * self._s[i2]
                    self._x[i2] = x2
                    del self._x[i1 + 1:i2]
                    del self._v[i1 + 1:i2]
                    del self._s[i1 + 1:i2]
                    return

    def add_slope(self, x1, x2, v1, s):
        """ Adds a piecewise linear step on an interval.

        Adds a piecewise linear step `f(x) = v1 + s * (x-x1)` to the invoking
        function on interval `[x1,x2)`.

        Args:
            x1 (int): Start of the interval.
            x2 (int): End of the interval.
            v1 (int or float): Added value at `x1`.
            s (int or float): Added function slope on interval `[x1,x2)`;
        """
        if v1 == 0 and s == 0:
            return
        assert s == 0 or not self.is_step_function, "Function not allowed on CpoStepFunction"
        assert is_int(x1) and is_int(x2), "Interval [x1,x2) must be integer"
        if x2 <= x1:
            return
        l = len(self._x)
        if l == 0:
            # Constant step function of value self._v0
            self._x.append(x1)
            self._v.append(self._v0 + v1)
            self._s.append(s)
            if x2 < INTERVAL_MAX:
                self._x.append(x2)
                self._v.append(self._v0)
                self._s.append(0)
            return
        if x1 < self._x[0]:
            # Adjust value of initial segment
            v0 = self._v0
            x0 = self._x[0]
            self._v0 += self._s0 * (x1 - self._x[0])
            # Add new segment
            self._x.insert(0, x1)
            self._v.insert(0, self._v0 + v1)
            self._s.insert(0, self._s0 + s)
            if x2 < x0:
                self._x.insert(1, x2)
                self._v.insert(1, v0 + (x2 - x0) * self._s0)
                self._s.insert(1, self._s0)
                return
            elif x0 <= x2:
                self.add_slope(x0, x2, v1 + (x0 - x1) * s, s)
                return
        else:
            # x1 >= self._x[0]:
            i1 = bisect.bisect_right(self._x, x1) - 1
            i2 = bisect.bisect_right(self._x, x2) - 1
            xcurr = self._x[i1]
            vcurr = self._v[i1]
            scurr = self._s[i1]
            if x1 > self._x[i1]:
                self._x.insert(i1 + 1, x1)
                self._v.insert(i1 + 1, vcurr + (x1 - self._x[i1]) * scurr + v1)
                self._s.insert(i1 + 1, scurr + s)
                i1 += 1
                i2 += 1
            else:
                assert x1 == self._x[i1]
                self._x[i1] = x1
                self._v[i1] = vcurr + v1
                self._s[i1] = scurr + s
            for i in range(i1 + 1, i2):
                self._v[i] = self._v[i] + v1 + (self._x[i] - x1) * s
                self._s[i] += s
            if i1 == i2:
                self._x.insert(i1 + 1, x2)
                self._v.insert(i1 + 1, vcurr + (x2 - xcurr) * scurr)
                self._s.insert(i1 + 1, scurr)
                return
            elif x2 > self._x[i2]:
                # Adjust beginning of first segment
                xcurr = self._x[i2]
                vcurr = self._v[i2]
                scurr = self._s[i2]
                self._v[i2] = self._v[i2] + v1 + (self._x[i2] - x1) * s
                self._s[i2] += s
                self._x.insert(i2 + 1, x2)
                self._v.insert(i2 + 1, vcurr + (x2 - xcurr) * scurr)
                self._s.insert(i2 + 1, scurr)
                return

    def _add(self, other):
        ov0 = other._v0
        os0 = other._s0
        ol = len(other._x)
        if ol == 0:
            # 'other' is a constant function of value ov0
            self._v0 += ov0
            for i in range(len(self._x)):
                self._v[i] += ov0
        else:
            ox0 = other._x[0]
            if ox0 == self._x[0]:
                self._v0 += ov0
                self._s0 += os0
            elif ox0 < self._x[0]:
                self._v0 = (ox0 - self._x[0]) * self._s0
                self._x.insert(0, ox0)
                self._v.insert(0, self._v0)
                self._s.insert(0, self._s0)
                self._v0 += ov0
                self._s0 += os0
            else:
                # self._x[0] < ox0
                self._s0 += os0
                self._v0 += ov0 + (self._x[0] - ox0) * os0
                self.add_slope(self._x[0], ox0, ov0 + (self._x[0] - ox0) * os0, os0)
            for i in range(ol - 1):
                # This could be optimized to avoid calling add_slope
                self.add_slope(other._x[i], other._x[i + 1], other._v[i], other._s[i])
            self.add_slope(other._x[ol - 1], INTERVAL_MAX, other._v[ol - 1], other._s[ol - 1])
        return self

    def _add_number(self, k):
        self._v0 += k
        for i in range(len(self._x)):
            self._v[i] += k
        return self

    def _mul(self, k):
        self._s0 *= k
        self._v0 *= k
        for i in range(len(self._x)):
            self._v[i] *= k
            self._s[i] *= k
        return self

    def __iadd__(self, other):
        if isinstance(other, CpoFunction):
            assert (self.is_segmented_function or not other.is_segmented_function)
            return self._add(other)
        elif is_number(other) or is_bool(other):
            return self._add_number(other)

    def __isub__(self, other):
        if isinstance(other, CpoFunction):
            assert (self.is_segmented_function or not other.is_segmented_function)
            tmp = other.copy()._mul(-1)
            return self._add(tmp)
        elif is_number(other) or is_bool(other):
            return self._add_number(-other)

    def __add__(self, other):
        if isinstance(other, CpoFunction):
            if self.is_segmented_function:
                return other.copy(Type_SegmentedFunction)._add(self)
            else:
                return other.copy()._add(self)
        elif is_number(other) or is_bool(other):
            return self.copy()._add_number(other)

    def __radd__(self, value):
        return self.__add__(value)

    def __sub__(self, other):
        if isinstance(other, CpoFunction):
            if self.is_segmented_function:
                return other.copy(Type_SegmentedFunction)._mul(-1)._add(self)
            else:
                return other.copy()._mul(-1)._add(self)
        elif is_number(other) or is_bool(other):
            return self.copy()._add_number(-other)

    def __neg__(self):
        return self.copy()._mul(-1)

    def __imul__(self, value):
        return self._mul(value)

    def __mul__(self, value):
        return self.copy()._mul(value)

    def __rmul__(self, value):
        return self.__mul__(value)

    def set_value(self, x1, x2, v):
        """ Sets the value of the function over an interval.

        Sets the value of the invoking function to be constant and equal to `v`
        on the interval `[x1,x2)`.

        Args:
            x1 (int): Start of the interval.
            x2 (int): End of the interval.
            v (int or float): Function value.
        """
        self.set_slope(x1, x2, v, 0)

    def add_value(self, x1, x2, v):
        """ Adds a constant to the value of the function over an interval.

        Adds constant value `v` to the value of the invoking function
        on the interval `[x1,x2)`.

        Args:
            x1 (int): Start of the interval.
            x2 (int): End of the interval.
            v (int or float): Added value.
        """
        self.add_slope(x1, x2, v, 0)

    def __pos__(self):
        return self


class CpoSegmentedFunction(CpoFunction):
    """ Class representing a segmented function.

    In CP Optimizer, piecewise linear functions are typically used in modeling a known function of time,
    for instance the cost that is incurred for completing an activity after a known date.

    A segmented function is a piecewise linear function defined on an interval [xmin, xmax)
    which is partitioned into segments such that over each segment, the function is linear.

    When two consecutive segments of the function are colinear, these segments are merged so that the function
    is always represented with the minimal number of segments.
    """

    def __init__(self, segment0=None, segments=None, name=None):
        """
        Args:
            segment0 (tuple): Initial segment of the function (slope, vright).
            segments (list): Segments of the function represented as a list of
                tuples (xleft, vleft, slope).
        """
        s0 = 0
        v0 = 0
        x = []
        v = []
        s = []
        if segment0 is not None:
            s0 = segment0[0]
            v0 = segment0[1]
        if segments is not None:
            for seg in segments:
                x.append(seg[0])
                v.append(seg[1])
                s.append(seg[2])
        super(CpoSegmentedFunction, self).__init__(Type_SegmentedFunction, s0=s0, v0=v0, x=x, v=v, s=s, name=name)

    def get_segment_list(self):
        """ Returns the list of segments of the function.
        """
        segments = [(self.s0, self.v0)]
        for i in range(len(self.x)):
            segments.append((self.x[i], self.v[i], self.s[i]))
        return segments

    def _get_expr_string(self):
        """ Get the string representing this expression, name excluded. """
        return "SegmentedFunction(...)"


class CpoStepFunction(CpoFunction):
    """ Class representing a step function.

    In CP Optimizer, stepwise functions are typically used to model the efficiency of a resource over time.

    A stepwise function is a special case of piecewise linear function where all slopes are equal to 0 and the
    domain and image of the function are integer.

    When two consecutive steps of the function have the same value, these steps are merged so that the function
    is always represented with the minimal number of steps.
    """
    def __init__(self, steps=None, name=None):
        """ Step function.

        Args:
            steps:  Function steps
            name: Function name
        """
        if steps is None or len(steps) == 0:
            super(CpoStepFunction, self).__init__(Type_StepFunction, name=name)
        else:
            x = []
            v = []
            if steps[0][0] == INTERVAL_MIN:
                v0 = steps[0][1]
            else:
                v0 = 0
                x.append(steps[0][0])
                v.append(steps[0][1])
            for s in steps[1:]:
                x.append(s[0])
                v.append(s[1])
            super(CpoStepFunction, self).__init__(typ=Type_StepFunction, s0=None, v0=v0, x=x, v=v, s=None, name=name)

    def get_step_list(self):
        """ Returns the list of steps of the function.
        """
        steps = []
        if self.v0 != 0:
            steps.append((INTERVAL_MIN, self.v0))
        for i in range(len(self.x)):
            steps.append((self.x[i], self.v[i]))
        return steps

    def _get_expr_string(self):
        """ Get the string representing this expression, name excluded. """
        return "StepFunction(...)"





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module implements appropriate software to solve a CPO model represented by a
:class:`docplex.cp.model.CpoModel` object.

It implements the following object classes:

 * :class:`CpoSolver` contains the public interface allowing to make solving requests with a model.
 * :class:`CpoSolverAgent` is an abstract class that is extended by the actual implementation(s) of
   the solving functions.

The :class:`CpoSolver` identifies and creates the required :class:`CpoSolverAgent` depending on the configuration
parameter *context.solver.agent' that contains the name of the agent to be used. This name is used to
access the configuration context *context.solver.<agent>* that contains the details about this agent.

For example, the default configuration refers to *docloud* as default solver agent, to solve model using *DOcplexcloud*
services. This means that at least following configuration elements must be set:
::

   context.solver.agent = 'docloud'
   context.solver.docloud.url = <URL of the service>
   context.solver.docloud.key = <Access key of the service>

The different methods that can be called on a CpoSolver object are:

 * :meth:`solve` simply solve the model and returns a solve result, if any.
   For convenience reason, this method is also directly available on the CpoModel object (:meth:`docplex.cp.model.CpoModel.solve`).
 * :meth:`search_next` and :meth:`end_search` allows to iterate on different solutions of the model.
 * :meth:`refine_conflict` calls the conflict refiner that identifies a minimal conflict for the infeasibility of
   the model.
 * :meth:`propagate` calls the propagation that communicates the domain reduction of a decision variable to
   all of the constraints that are stated over this variable.

Except :meth:`solve`, these functions are only available with a local solver with release greater or equal to 12.7.0.0
When a method is not available, an exception *CpoNotSupportedException* is raised.

If the methods :meth:`search_next` and :meth:`end_search` are available in the underlying solver agent,
the :class:`CpoSolver` object acts as an iterator. All solutions are retrieved using a loop like:
::

   solver = CpoSolver(mdl)
   for sol in solver:
       sol.write()

A such solution iteration can be interrupted at any time by calling end_search() that returns
a fail solution including the last solve status.


Detailed description
--------------------
"""

import docplex.cp.config as config
from docplex.cp.cpo.cpo_compiler import CpoCompiler
from docplex.cp.solution import *
from docplex.cp.solver.solver_listener import CpoSolverListener
from docplex.cp.solver.cpo_callback import CpoCallback

import time, importlib, inspect
import traceback
import threading


###############################################################################
##  Public constants
###############################################################################

# Solver statuses
STATUS_IDLE              = "Idle"             # Solver created but inactive
STATUS_RELEASED          = "Released"         # Solver stopped with resources released.
STATUS_ABORTED           = "Aborted"          # Solver has been aborted.
STATUS_SOLVING           = "SolveRunning"     # Simple solve in progress
STATUS_SEARCH_WAITING    = "SearchWaiting"    # Search started or waiting to call next
STATUS_SEARCH_RUNNING    = "NextRunning"      # Search of next solution in progress
STATUS_REFINING_CONFLICT = "RefiningConflict" # Solver refine conflict in progress
STATUS_PROPAGATING       = "Propagating"      # Propagation in progress
STATUS_RUNNING_SEEDS     = "RunningSeeds"     # Run seeds in progress

# Set of statuses that end solver
_ENDING_STATUSES = frozenset((STATUS_RELEASED, STATUS_ABORTED))


###############################################################################
##  Public classes
###############################################################################

class CpoSolverAgent(object):
    """ This class is an abstract class that must be extended by every solver agent that intend
    to be called by :class:`CpoSolver` to solve a CPO model.
    """
    __slots__ = ('name',             # Agent name
                 'solver',           # Parent solver
                 'model',            # Source model
                 'params',           # Solver parameters
                 'context',          # Solve context
                 'last_json_result', # String of the last received JSON result
                 'rename_map',       # Map of renamed variables. Key is new name, value is original name
                 'version_info',     # Solver version information (dict). None if unknown.
                 'process_infos',    # Processing information
                 'log_output',       # Log output stream
                 'log_print',        # Print log indicator
                 'log_data',         # Log data buffer (list of strings)
                 'log_enabled',      # Global log enabled indicator
                 'expr_map',         # Map of expressions to rebuild result
                )

    def __init__(self, solver, params, context):
        """ Constructor

        Args:
            solver:   Parent solver, object of type CpoSolver
            params:   Solving parameters, object of type CpoParameters
            context:  Solver agent context
        Raises:
            CpoException if agent can not be created properly.
        """
        super(CpoSolverAgent, self).__init__()
        self.solver = solver
        self.model = solver.get_model()
        self.params = params
        self.context = context
        self.last_json_result = None
        self.rename_map = None
        self.version_info = None
        self.process_infos = CpoProcessInfos()

        # Initialize log
        self.log_output = context.get_log_output()
        self.log_print = context.trace_log and (self.log_output is not None)
        self.log_data = [] if context.add_log_to_solution else None
        self.log_enabled = self.log_print or (self.log_data is not None)


    def solve(self):
        """ Solve the model

        Returns:
            Model solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def start_search(self):
        """ Start a new search. Solutions are retrieved using method search_next().

        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def search_next(self):
        """ Search the next available solution.

        Returns:
            Next solve result,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def end_search(self):
        """ End current search.

        Returns:
            Last (fail) solve result with last solve information,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def abort_search(self):
        """ Abort current search.
        This method is designed to be called by a different thread than the one currently solving.

        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def refine_conflict(self):
        """ This method identifies a minimal conflict for the infeasibility of the current model.

        Given an infeasible model, the conflict refiner can identify conflicting constraints and variable domains
        within the model to help you identify the causes of the infeasibility.
        In this context, a conflict is a subset of the constraints and/or variable domains of the model
        which are mutually contradictory.
        Since the conflict is minimal, removal of any one of these constraints will remove that
        particular cause for infeasibility.
        There may be other conflicts in the model; consequently, repair of a given conflict
        does not guarantee feasibility of the remaining model.

        Returns:
            Conflict result,
            object of class :class:`~docplex.cp.solution.CpoRefineConflictResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def propagate(self):
        """ This method invokes the propagation on the current model.

        Constraint propagation is the process of communicating the domain reduction of a decision variable to
        all of the constraints that are stated over this variable.
        This process can result in more domain reductions.
        These domain reductions, in turn, are communicated to the appropriate constraints.
        This process continues until no more variable domains can be reduced or when a domain becomes empty
        and a failure occurs.
        An empty domain during the initial constraint propagation means that the model has no solution.

        The result is a object of class CpoSolveResult, the same than the one returned by solve() method.
        However, in this case, variable domains may not be completely defined.

        Returns:
            Propagation result,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def run_seeds(self, nbrun):
        """ This method runs *nbrun* times the CP optimizer search with different random seeds
        and computes statistics from the result of these runs.

        Result statistics are displayed on the log output that should be activated.
        If the appropriate configuration variable *context.solver.add_log_to_solution* is set to True (default),
        log is also available in the *CpoRunResult* result object, accessible as a string using the method
        :meth:`~docplex.cp.solution.CpoRunResult.get_solver_log`

        Each run of the solver is stopped according to single solve conditions (TimeLimit for example).
        Total run time is then expected to take *nbruns* times the duration of a single run.

        Args:
            nbrun: Number of runs with different seeds.
        Returns:
            Run result, object of class :class:`~docplex.cp.solution.CpoRunResult`.
        Raises:
            CpoNotSupportedException: method not available in this solver agent.
        """
        self._raise_not_supported()


    def end(self):
        """ End solver agent and release all resources.
        """
        self.solver = None
        self.model = None
        self.params = None
        self.context = None


    def _get_cpo_model_string(self):
        """ Get the CPO model as a string, according to configuration

        Return:
            String containing the CPO model in CPO file format
        """
        # Build string
        ctx = self.context
        stime = time.time()
        cplr = CpoCompiler(self.model, params=self.params, context=ctx.get_root())
        cpostr = cplr.get_as_string()
        self.expr_map = cplr.get_expr_map()
        self.process_infos[CpoProcessInfos.MODEL_COMPILE_TIME] = time.time() - stime
        self.process_infos[CpoProcessInfos.MODEL_DATA_SIZE] = len(cpostr)

        # Trace CPO model if required
        lout = ctx.get_log_output()
        if lout and ctx.trace_cpo:
            stime = time.time()
            lout.write("Model '" + str(self.model.get_name()) + "' in CPO format:\n")
            lout.write(cpostr)
            lout.write("\n")
            self.model.write_information(lout)
            lout.write("\n")
            lout.flush()
            self.process_infos.incr(CpoProcessInfos.MODEL_DUMP_TIME, time.time() - stime)

        # Dump in dump directory if required
        if ctx.model.dump_directory:
            stime = time.time()
            make_directories(ctx.model.dump_directory)
            mname = self.model.get_name()
            if mname is None:
                mname = "Anonymous"
            else:
                # Remove special characters introduced by Jupyter
                mname = mname.replace('<', '').replace('>', '')
            file = ctx.model.dump_directory + "/" + mname + ".cpo"
            with utils.open_utf8(file, 'w') as f:
                f.write(cpostr)
            self.process_infos.incr(CpoProcessInfos.MODEL_DUMP_TIME, time.time() - stime)

        # Return
        return cpostr


    def _add_log_data(self, data):
        """ Add new log data
        Args:
            data:  Data to log (String)
        """
        self.solver._notify_new_log(data)
        if self.log_enabled:
            if self.log_print:
                self.log_output.write(data)
                self.log_output.flush()
            if self.log_data is not None:
                self.log_data.append(data)
        # Update statistics
        self.process_infos.incr(CpoProcessInfos.TOTAL_LOG_DATA_SIZE, len(data))



    def _set_last_json_result_string(self, json):
        """ Set the string containing last received JSON result

        Args:
            json: JSON result string
        """
        self.last_json_result = json


    def _get_last_json_result_string(self):
        """ Get the string containing last received JSON result

        Return:
            Last JSON result string, None if none
        """
        return self.last_json_result


    def _create_result_object(self, rclass, jsol=None):
        """ Create a new result object and fill it with necessary data
        Args:
            rclass:            Result object class
            jsol (optional):   JSON solution string
        Returns:
            New result object preinitialized
        """
        res = rclass(self.model)
        res.process_infos.update(self.process_infos)

        # Process JSON solution
        self.context.log(3, "JSON data:\n", jsol)
        self.last_json_result = jsol

        # Parse JSON solution
        if jsol:
            # Parse JSON
            stime = time.time()
            jsol = parse_json_string(jsol)
            res.process_infos.incr(CpoProcessInfos.TOTAL_JSON_PARSE_TIME, time.time() - stime)
            # Build result structure
            res._add_json_solution(jsol, self.expr_map)

        # Process Log
        if self.log_data is not None:
            res._set_solver_log(''.join(self.log_data))
            self.log_data = []
        return res


    def _raise_not_supported(self):
        """ Raise an exception indicating that the calling method is not supported.
        """
        raise CpoNotSupportedException("Method '{}' is not available in solver agent '{}' ({})."
                                       .format(inspect.stack()[1][3], self.context.agent, type(self)))


class CpoSolver(object):
    """ This class represents the public API of the object allowing to solve a CPO model.

    It create the appropriate :class:`CpoSolverAgent` that actually implements solving functions, depending
    on the value of the configuration parameter *context.solver.agent*.
    """
    __slots__ = ('model',        # Model to solve
                 'context',      # Solving context
                 'agent',        # Solver agent
                 'status',       # Current solver status
                 'last_result',  # Last returned solution
                 'listeners',    # List of solve listeners
                 'status_lock',  # Lock protecting status change
                 'callbacks',    # List of CPO solver callbacks
                )

    def __init__(self, model, **kwargs):
        """ Constructor

        All necessary solving parameters are taken from the solving context that is constructed from the following list
        of sources, each one overwriting the previous:

           - the parameters that are set in the model itself,
           - the default solving context that is defined in the module :mod:`~docplex.cp.config`
           - the user-specific customizations of the context that may be defined (see :mod:`~docplex.cp.config` for details),
           - the optional arguments of this method.

        Args:
            context (Optional): Complete solving context.
                                If not given, solving context is the default one that is defined in the module
                                :mod:`~docplex.cp.config`.
            params (Optional):  Solving parameters (object of class :class:`~docplex.cp.parameters.CpoParameters`)
                                that overwrite those in the solving context.
            url (Optional):     URL of the DOcplexcloud service that overwrites the one defined in the solving context.
            key (Optional):     Authentication key of the DOcplexcloud service that overwrites the one defined in
                                the solving context.
            (param) (Optional): Any individual solving parameter as defined in class :class:`~docplex.cp.parameters.CpoParameters`
                               (for example *TimeLimit*, *Workers*, *SearchType*, etc).
            (others) (Optional): Any leaf attribute with the same name in the solving context
                                (for example *agent*, *trace_log*, *trace_cpo*, etc).
            (listeners) (Optional): List of solution listeners
        """
        super(CpoSolver, self).__init__()
        self.agent = None
        self.last_result = None
        self.status = STATUS_IDLE
        self.status_lock = threading.Lock()
        self.listeners = []
        self.callbacks = []

        # Build effective context from args
        context = config._get_effective_context(**kwargs)

        # If defined, limit the number of threads
        mxt = context.solver.max_threads
        if isinstance(mxt, int):
            # Maximize number of workers
            nbw = context.params.Workers
            if (nbw is None) or (nbw > mxt):
                context.params.Workers = mxt
                print("WARNING: Number of workers has been reduced to " + str(mxt) + " to comply with platform limitations.")

        # Save attributes
        self.model = model
        self.context = context

        # Determine appropriate solver agent
        self.agent = self._get_solver_agent()

        # Add configured default listeners if any
        # Note: calling solver_created() is not required as it is done by add_listener().
        lstnrs = context.solver.listeners
        if lstnrs is not None:
            if is_array(lstnrs):
                for lstnr in lstnrs:
                    self._add_listener_from_class(lstnr)
            else:
                self._add_listener_from_class(lstnrs)


    def __iter__(self):
        """  Define solver as an iterator """
        return self


    def __del__(self):
        # End solver
        self.end()


    def __enter__(self):
        # For usage in with
        return self


    def __exit__(self, exception_type, exception_value, traceback):
        # End solver
        self.end()


    def get_model(self):
        """ Returns the model solved by this solver.

        Returns:
            Model solved by this solver
        """
        return self.model


    def get_model_format_version(self):
        """ If defined, returns the format version of the model.

        Returns:
            Model format version, None if not defined.
        """
        return None if self.model is None else self.model.get_format_version()


    def solve(self):
        """ Solve the model

        This function solves the model using CP Optimizer's built-in strategy.
        The built-in strategy is determined by setting the parameter SearchType (see docplex.cp.parameters).
        If the model contains an objective, then the optimal solution with respect to the objective will be calculated.
        Otherwise, a solution satisfying all problem constraints will be calculated.

        The function returns an object of the class CpoSolveResult (see docplex.cp.solution) that contains the solution
        if exists, plus different information on the solving process.

        If the context parameter *solve_with_start_next* (or config parameter *context.solver.solve_with_start_next*)
        is set to True, the call to solve() is replaced by loop start/next which returns the last solution found.
        If a solver listener has been added to the solver, it is warned of all intermediate solutions.

        Returns:
            Solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoException: (or derived) if error.
        """
        # Check solve with start/next
        if self.context.solver.solve_with_start_next:
            return self._solve_with_start_next()

        # Notify listeners
        for lstnr in self.listeners:
            lstnr.start_solve(self)

        # Solve model
        stime = time.time()
        self._set_status(STATUS_SOLVING)
        try:
            msol = self.agent.solve()
        except Exception as e:
            # Check if aborted in the mean time
            if self._check_status_aborted():
                return self.last_result
            if self.context.log_exceptions:
                traceback.print_exc()
            raise e
        self._set_status(STATUS_IDLE)
        stime = time.time() - stime
        self.context.solver.log(1, "Model '", self.model.get_name(), "' solved in ", round(stime, 2), " sec.")
        msol.process_infos[CpoProcessInfos.SOLVE_TOTAL_TIME] = stime

        # Set solve time in solution if not done
        if msol.get_solve_time() == 0:
            msol._set_solve_time(stime)

        # Store last solution
        self.last_result = msol

        # Notify listeners
        for lstnr in self.listeners:
            lstnr.result_found(self, msol)
        for lstnr in self.listeners:
            lstnr.end_solve(self)

        # Return solution
        return msol
        
     
    def search_next(self):
        """ Get the next available solution.

        This function is available only with local CPO solver with release number greater or equal to 12.7.0.

        Returns:
            Next solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoNotSupportedException: if method not available in the solver agent.
        """
        # Initiate search if needed
        if self.status == STATUS_IDLE:
            # Notify listeners about start of search
            self.agent.start_search()
            self._set_status(STATUS_SEARCH_WAITING)
            for lstnr in self.listeners:
                lstnr.start_solve(self)

        # Check if status is aborted in the mean time (may be caused by listener)
        if self._check_status_aborted():
            return self.last_result

        self._check_status(STATUS_SEARCH_WAITING)

        # Search next
        stime = time.time()
        self._set_status(STATUS_SEARCH_RUNNING)
        try:
            msol = self.agent.search_next()
        except BaseException as e:
            sys.stdout.flush()
            # Check if aborted in the mean time
            if self._check_status_aborted():
                return self.last_result
            if self.context.log_exceptions:
                traceback.print_exc()
            raise e
        self._set_status(STATUS_SEARCH_WAITING)
        stime = time.time() - stime
        self.context.solver.log(1, "Model '", self.model.get_name(), "' next solution in ", round(stime, 2), " sec.")

        # Set solve time in solution if not done
        if msol.get_solve_time() == 0:
            msol._set_solve_time(stime)

        # Store last solution
        self.last_result = msol

        # Notify listeners
        for lstnr in self.listeners:
            lstnr.result_found(self, msol)

        # Return solution
        return msol


    def end_search(self):
        """ End current search.

        This function is available only with local CPO solver with release number greater or equal to 12.7.0.

        Returns:
            Last (fail) model solution with last solve information,
            object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoNotSupportedException: if method not available in the solver agent.
        """
        if self.status == STATUS_RELEASED or self.status == STATUS_ABORTED:
           return self.last_result
        self._check_status(STATUS_SEARCH_WAITING)
        msol = self.agent.end_search()
        self._set_status(STATUS_IDLE)
        self.last_result = msol
        for lstnr in self.listeners:
            lstnr.end_solve(self)
        return msol


    def refine_conflict(self):
        """ This method identifies a minimal conflict for the infeasibility of the current model.

        Given an infeasible model, the conflict refiner can identify conflicting constraints and variable domains
        within the model to help you identify the causes of the infeasibility.
        In this context, a conflict is a subset of the constraints and/or variable domains of the model
        which are mutually contradictory.
        Since the conflict is minimal, removal of any one of these constraints will remove that
        particular cause for infeasibility.
        There may be other conflicts in the model; consequently, repair of a given conflict
        does not guarantee feasibility of the remaining model.

        Conflict refiner is controlled by the following parameters (that can be set at CpoSolver creation):

         * ConflictRefinerBranchLimit
         * ConflictRefinerFailLimit
         * ConflictRefinerIterationLimit
         * ConflictRefinerOnVariables
         * ConflictRefinerTimeLimit

        that are described in module :mod:`docplex.cp.parameters`.

        Note that the general *TimeLimit* parameter is used as a limiter for each conflict refiner iteration, but the
        global limitation in time must be set using *ConflictRefinerTimeLimit* that is infinite by default.

        This function is available only with local CPO solver with release number greater or equal to 12.7.0.

        Returns:
            List of constraints that cause the conflict,
            object of class :class:`~docplex.cp.solution.CpoRefineConflictResult`.
        Raises:
            CpoNotSupportedException: if method not available in the solver agent.
        """
        # Start refine conflict
        self._check_status(STATUS_IDLE)
        self._set_status(STATUS_REFINING_CONFLICT)
        for lstnr in self.listeners:
            lstnr.start_refine_conflict(self)

        # Start refine conflict
        msol = self.agent.refine_conflict()
        for lstnr in self.listeners:
            lstnr.conflict_found(self, msol)

        # End refine conflict
        self._set_status(STATUS_IDLE)
        for lstnr in self.listeners:
            lstnr.end_refine_conflict(self)


        return msol


    def propagate(self):
        """ This method invokes the propagation on the current model.

        Constraint propagation is the process of communicating the domain reduction of a decision variable to
        all of the constraints that are stated over this variable.
        This process can result in more domain reductions.
        These domain reductions, in turn, are communicated to the appropriate constraints.
        This process continues until no more variable domains can be reduced or when a domain becomes empty
        and a failure occurs.
        An empty domain during the initial constraint propagation means that the model has no solution.

        The result is a object of class CpoSolveResult, the same than the one returned by solve() method.
        However, variable domains may not be completely defined.

        This function is available only with local CPO solver with release number greater or equal to 12.7.0.

        Returns:
            Propagation result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        Raises:
            CpoNotSupportedException: method not available in configured solver agent.
        """
        self._check_status(STATUS_IDLE)
        self._set_status(STATUS_PROPAGATING)
        psol = self.agent.propagate()
        self._set_status(STATUS_IDLE)
        return psol


    def run_seeds(self, nbrun):
        """ This method runs *nbrun* times the CP optimizer search with different random seeds
        and computes statistics from the result of these runs.

        Result statistics are displayed on the log output that should be activated.
        If the appropriate configuration variable *context.solver.add_log_to_solution* is set to True (default),
        log is also available in the *CpoRunResult* result object, accessible as a string using the method
        :meth:`~docplex.cp.solution.CpoRunResult.get_solver_log`

        Each run of the solver is stopped according to single solve conditions (TimeLimit for example).
        Total run time is then expected to take *nbruns* times the duration of a single run.

        Args:
            nbrun: Number of runs with different seeds.
        Returns:
            Run result, object of class :class:`~docplex.cp.solution.CpoRunResult`.
        Raises:
            CpoNotSupportedException: method not available in configured solver agent.
        """
        self._check_status(STATUS_IDLE)
        self._set_status(STATUS_RUNNING_SEEDS)
        rsol = self.agent.run_seeds(nbrun)
        self._set_status(STATUS_IDLE)
        return rsol


    def get_last_solution(self):
        """ Get the last result returned by this solver

        DEPRECATED. Use get_last_result instead.

        Returns:
            Solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        """
        return self.last_result


    def get_last_result(self):
        """ Get the last result returned by this solver

        Returns:
            Solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        """
        return self.last_result


    def abort_search(self):
        # Abort current search if any. Temporarily, end the process
        self._set_status(STATUS_ABORTED)
        agt = self.agent
        self.agent = None
        if agt is not None:
            agt.abort_search()


    def end(self):
        # End this solver and release associated resources
        agt = self.agent
        self.agent = None
        self._set_status(STATUS_RELEASED)
        if agt is not None:
            agt.end()


    def next(self):
        # """ For solution iteration, get the next available solution.
        #
        # This function is available only with local CPO solver with release number greater or equal to 12.7.0.
        #
        # Returns:
        #     Next solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        # """
        # Get next solution
        msol = self.search_next()
        if msol:
            return msol
        self.end_search()
        raise StopIteration()


    def __next__(self):
        # """ Get the next available solution (same as next() for compatibility with Python 3)
        #
        # This function is available only with local CPO solver with release number greater or equal to 12.7.0.
        #
        # Returns:
        #     Next solve result, object of class :class:`~docplex.cp.solution.CpoSolveResult`.
        # """
        return self.next()


    def add_listener(self, lstnr):
        """ Add a solver listener.

        A solver listener is an object extending the class :class:`~docplex.cp.solver.solver_listener.CpoSolverListener`
        which provides multiple functions that are called to notify about the different solving steps.

        Args:
            lstnr:  Solver listener
        """
        assert isinstance(lstnr, CpoSolverListener), \
            "Listener should be an object of class docplex.cp.solver.solver_listener.CpoSolverListener"
        self.listeners.append(lstnr)
        # Notify listener
        lstnr.solver_created(self)


    def _add_listener_from_class(self, lstnr):
        """ Add a solver listener from its class (instance is created).

        Args:
            lstnr:  Solver listener class, or string identifying the class
        """
        if is_string(lstnr):
            # Get listener class from string
            try:
                lclass = utils.get_module_element_from_path(lstnr)
            except Exception as e:
                raise CpoException("Unable to retrieve solver listener class '{}': {}".format(lstnr, e))
            if not inspect.isclass(lclass):
                raise CpoException("Solver listener '{}' is not a class.".format(lstnr))
            if not issubclass(lclass, CpoSolverListener):
                raise CpoException("Solver listener class '{}' should extend CpoSolverListener.".format(lstnr))
        else:
            # Listener is assumed to directly be a class
            lclass = lstnr
            if not inspect.isclass(lclass):
                raise CpoException("Solver listener '{}' is not a class.".format(lclass))
            if not issubclass(lclass, CpoSolverListener):
                raise CpoException("Solver listener class '{}' should extend CpoSolverListener.".format(lclass))
        # Add listener
        self.add_listener(lclass())


    def remove_listener(self, lstnr):
        """ Remove a solver listener previously added with :meth:`~docplex.cp.solver.solver.CpoSolver.add_listener`.

        Args:
            lstnr:  Listener to remove.
        """
        self.listeners.remove(lstnr)


    def add_callback(self, cback):
        """ Add a CPO solver callback.

        A solver callback is an object extending the class :class:`~docplex.cp.solver.cpo_callback.CpoCallback`
        which provides multiple functions that are called to notify about the different solving steps.

        Args:
            cback:  Solver callback
        """
        assert isinstance(cback, CpoCallback), \
            "CPO callback should be an object of class docplex.cp.solver.cpo_callback.CpoCallback"
        self.callbacks.append(cback)


    def remove_callback(self, cback):
        """ Remove a CPO solver callback. previously added with :meth:`~docplex.cp.solver.solver.CpoSolver.add_callback`.

        Args:
            cback:  Callback to remove.
        """
        self.callbacks.remove(cback)


    def _set_status(self, status):
        """ Change solve status, only if allowed

        Args:
            status: New solve status
        """
        with self.status_lock:
            if (status in _ENDING_STATUSES) or (not self.status in _ENDING_STATUSES):
                self.status = status


    def _notify_new_log(self, data):
        """ Notify new log data (called by agent)

        Args:
            data: Log data as a string
        """
        # Notify listeners
        for lstnr in self.listeners:
            lstnr.new_log_data(self, data)


    def _notify_callback_event(self, event, data):
        """ Notify a CPO callback event (called by agent)

        Args:
            event:  Event id
            data:   JSON document associated to this event
        """
        # Notify callbacks
        for cback in self.callbacks:
            cback.invoke(self, event, data)


    def _solve_with_start_next(self):
        """ Solve the model using a start/next loop instead of standard solve.

        Return:
            Last solve result
        """
        # Loop on all solutions
        last_sol = None
        while True:
            # Search for next solution
            msol = self.search_next()

            # Check successful search
            if msol:
                last_sol = msol
            else:
                break

        # Process end of search
        if last_sol is None:
            last_sol = msol
        else:
            # Merge last valid solution with last solve infos
            last_sol.solver_infos = msol.solver_infos
        self.end_search()
        return last_sol


    def _check_status(self, ests):
        """ Throws an exception if solver status is not the expected one

        Args:
            ests:  Expected status, or list of expected statuses
        Raise:
            CpoException if solver is not in the right status
        """
        if self.status != ests:
           raise CpoException("Unexpected solver status. Should be '{}' instead of '{}'".format(ests, self.status))


    def _check_status_aborted(self):
        """ Check if the solve status has been changed to aborted by another thread or a listener.
        If so, an Aborted solve result is stored in last_result.

        Returns: true if status was aborted, false otherwise.
        """
        if self.status != STATUS_ABORTED:
            return False

        self._set_status(STATUS_RELEASED)
        self.last_result = self._create_solution_aborted()
        for lstnr in self.listeners:
            lstnr.result_found(self, self.last_result)
        for lstnr in self.listeners:
            lstnr.end_solve(self)
        return True


    def _create_solution_aborted(self):
        """ Create an empty solution with aborted status
        """
        res = CpoSolveResult(self.model)
        res.solve_status = SOLVE_STATUS_JOB_ABORTED
        res.fail_status = FAIL_STATUS_ABORT
        res.search_status = SEARCH_STATUS_STOPPED
        res.stop_cause = STOP_CAUSE_ABORT
        return res


    def _get_solver_agent(self):
        """ Get the solver agent instance that is used to solve the model.

        Returns:
            Solver agent instance
        Raises:
            CpoException:  Agent creation error
        """
        # Determine selectable agent(s)
        sctx = self.context.solver

        alist = sctx.agent
        if alist is None:
            alist = 'docloud'
        elif not (is_string(alist) or is_array(alist)):
            raise CpoException("Agent identifier in config.context.solver.agent should be a string or a list of strings.")

        # Create agent
        if is_string(alist):
            aname = alist
            agent = self._create_solver_agent(alist)
        else:
            # Search first available agent in the list
            agent = None
            aname = None
            errors = []
            for aname in alist:
                try:
                    agent = self._create_solver_agent(aname)
                    break
                except Exception as e:
                    errors.append((aname, str(e)))
                # Agent not found
                errstr = ', '.join(a + ": " + str(e) for (a, e) in errors)
                raise CpoException("Agent creation error: " + errstr)

        # Log solver agent
        sctx.log(1, "Solve model '", self.model.get_name(), "' with agent '", aname, "'")
        agent.process_infos[CpoProcessInfos.SOLVER_AGENT] = aname
        return agent


    def _create_solver_agent(self, aname):
        """ Create a new solver agent from its name.

        Args:
            name: Name of the agent
        Returns:
            Solver agent instance
        Raises:
            CpoException: Agent creation error
        """
        # Get agent context
        sctx = self.context.solver.get(aname)
        if not isinstance(sctx, Context):
            raise CpoException("Unknown solving agent '" + aname + "'. Check config.context.solver.agent parameter.")
        if sctx.is_log_enabled(3):
            sctx.log(3, "Context for solving agent '", aname, "':")
            sctx.write(out=sctx.get_log_output())
        cpath = sctx.class_name
        if cpath is None:
            raise CpoException("Solving agent '" + aname + "' context should contain an attribute 'class_name'")

        # Retrieve solver agent class
        try:
            sclass = utils.get_module_element_from_path(cpath)
        except Exception as e:
            raise CpoException("Unable to retrieve solver agent class '{}': {}".format(cpath, e))
        if not inspect.isclass(sclass):
            raise CpoException("Solver agent '{}' is not a class.".format(cpath))
        if not issubclass(sclass, CpoSolverAgent):
            raise CpoException("Solver agent class '{}' should extend CpoSolverAgent.".format(cpath))

        # Create agent instance
        agent = sclass(self, sctx.params, sctx)
        return agent


###############################################################################
##  Public functions
###############################################################################

def get_version_info():
    """ If the solver agent defined in the configuration enables this function,
    this method returns solver version information.

    This method creates a CP solver to retrieve this information, and end it immediately.
    It returns a dictionary with various information, as in the following example:
    ::
    {
       "ProxyVersion" : 5,
       "SourceDate" : "Sep 12 2017",
       "SolverVersion" : "12.8.0.0",
       "IntMin" : -9007199254740991,
       "IntMax" : 9007199254740991,
       "IntervalMin" : -4503599627370494,
       "IntervalMax" : 4503599627370494,
    }

    Returns:
        Solver information dictionary, or None if not available.
    """
    from docplex.cp.model import CpoModel
    try:
        with CpoSolver(CpoModel()) as slvr:
            return slvr.agent.version_info
    except:
        pass
    return None


def get_solver_version():
    """ If the solver agent defined in the configuration enables this function,
    this method returns solver version number.

    Returns:
        Solver version string, or None if not available.
    """
    vinfo = get_version_info()
    return vinfo.get('SolverVersion') if vinfo else None


###############################################################################
##  Private Functions
###############################################################################

def _get_solver_agent_class(aname, sctx):
    """ Get a solver agent class from its name

    Args:
        aname:  Solver agent name
        sctx:   Candidate solver context
    Returns:
        Solver agent class
    """
    # Check for solver agent context
    if not isinstance(sctx, Context):
        raise CpoException("Unknown solving agent '" + aname + "'. Check config.context.solver.agent parameter.")
    cpath = sctx.class_name
    if cpath is None:
        raise CpoException("Solving agent '" + aname + "' context does not contain attribute 'class_name'")

    # Split class name
    pnx = cpath.rfind('.')
    if pnx < 0:
        raise CpoException("Invalid class name '" + cpath + "' for solving agent '" + aname + "'. Should be <package>.<module>.<class>.")
    mname = cpath[:pnx]
    cname = cpath[pnx + 1:]

    # Load module
    try:
        module = importlib.import_module(mname)
    except Exception as e:
        raise CpoException("Module '" + mname + "' import error: " + str(e))

    # Create and check class
    sclass = getattr(module, cname, None)
    if sclass is None:
        raise CpoException("Module '" + mname + "' does not contain a class '" + cname + "'")
    if not inspect.isclass(sclass):
        raise CpoException("Agent class '" + cpath + "' is not a class.")
    if not issubclass(sclass, CpoSolverAgent):
        raise CpoException("Solver agent class '" + cpath + "' does not extend CpoSolverAgent.")

    # Return
    return sclass

def _replace_names_in_json_dict(jdict, renmap):
    """ Replace keys that has been renamed in a JSON result directory
    Args:
        jdict:  Json result dictionary
        renmap: Renaming map, key is name to replace, value is name to use instead
    """
    if jdict:
        for k in list(jdict.keys()):
            nk = renmap.get(k)
            if nk:
                jdict[nk] = jdict[k]
                del jdict[k]




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
This module contains the basic classes representing the expressions required to
describe a constraint programming model.

In particular, it defines the following classes:

 * :class:`CpoExpr`: the root class of each model expression node,
 * :class:`CpoIntVar`: representation of an integer variable,
 * :class:`CpoIntervalVar`: representation of an interval variable,
 * :class:`CpoSequenceVar`: representation of a sequence variable,
 * :class:`CpoTransitionMatrix`: representation of a transition matrix,
 * :class:`CpoStateFunction`: representation of a state function.

**None of these classes should be created explicitly.**
There are various factory functions to do so, such as:

 * :meth:`integer_var`, :meth:`integer_var_list`, :meth:`integer_var_dict` to create integer variable(s),
 * :meth:`binary_var`, :meth:`binary_var_list`, :meth:`binary_var_dict` to create integer variable(s) with value in [0..1],
 * :meth:`interval_var`, :meth:`interval_var_list`, :meth:`interval_var_dict` to create interval variable(s),
 * :meth:`sequence_var` to create a sequence variable,
 * :meth:`transition_matrix` to create a transition matrix,
 * :meth:`tuple_set` to create a tuple set,
 * :meth:`state_function` to create a state function.

Moreover, some automatic conversions are also provided to generate CP Optimizer objects from Python objects.
For example:

 * an integer is converted into CP constant,
 * an iterator of objects is converted in CP array of objects of the corresponding type,
 * a tuple of tuples is converted into tuple set.


Detailed description
--------------------
"""

from docplex.cp.utils import *
from docplex.cp.catalog import *
from docplex.cp.config import context

import threading
import warnings


###############################################################################
## Constants
###############################################################################

INT_MAX = (2**53 - 1)  # (2^53 - 1) for 64 bits, (2^31 - 1) for 32 bits
""" Maximum integer value. """

INT_MIN = -INT_MAX
""" Minimum integer value. """

DEFAULT_INTEGER_VARIABLE_DOMAIN = ((INT_MIN, INT_MAX),)
""" Default integer variable domain """

# Domain for binary variables
BINARY_DOMAIN = ((0, 1),)

# Floating point precision to verify equality of floats
_FLOATING_POINT_PRECISION = 1e-9


###############################################################################
## Public expression classes
###############################################################################

class CpoExpr(object):
    """ This class is an abstract class that represents any CPO expression node.

    It does not contain links to children expressions that are implemented in extending classes.
    However, method allowing to access to children is provided with default return value.
    """
    __slots__ = ('type',             # Expression result type
                 'name',             # Name of the expression (None if none)
                 'priority',         # Operation priority
                 'children',         # List of children, empty tuple if none
                )

    # To force possible numpy operators overloading to get CPO expressions as main operand
    __array_priority__ = 100

    def __init__(self, type, name):
        """ Constructor:

        Args:
            type:   Expression type, object of class CpoType.
            name:   Expression name.
        """
        super(CpoExpr, self).__init__()
        self.type = type
        self.name = name
        self.priority = -1
        self.children = ()


    def __hash__(self):
        """ Redefinition of hash function (mandatory for Python 3)
        """
        return int(id(self) / 16)


    '''
    # These functions are required with standard 'pickle'
    # But DO NOT activate them with cloudpickle
    def __getstate__(self):
        """ Build a picklable object from this object
        """
        return dict((k, getattr(self, k, None)) for k in self.__slots__)

    def __setstate__(self, data):
        """ Fill object from its pickle form
        """
        for (k, v) in data.iteritems():
            setattr(self, k, v)
    '''


    def set_name(self, name):
        """ Set the name of the expression.

        This method returns this expression. It enables fluent calls such as: mdl.add((a + b == 1).set_name("myname"))

        Args:
            name: Expression name, possibly None.
        Returns:
            This expression
        """
        assert (name is None) or is_string(name), "Argument 'name' should be a string or None, not '{}' of type {}".format(name, type(name))
        if name is not None:
            name = make_unicode(name)
        self.name = name
        return self


    def get_name(self):
        """ Get the name of the expression.

        Returns:
            Name of the expression, None if not defined
        """
        return self.name


    def has_name(self):
        """ Check if this expression has a name

        Returns:
            True if this expression has a name, False otherwise.
        """
        return self.name is not None


    def is_type(self, xtyp):
        """ Check if the type of this expression is a given one

        Args:
            xtyp:  Expected type
        Returns:
            True if expression type is the expected one
        """
        return self.type == xtyp


    def is_kind_of(self, tp):
        """ Checks if the type of this expression type is compatible with another type.

        Args:
            tp: Other type to check.
        Returns:
           True if this expression type is a kind of tp.
        """
        # Check if required type is the same
        return self.type.is_kind_of(tp)


    def is_variable(self):
        """ Check if this expression is a variable

        Returns:
            True if this expression is a variable
        """
        return False


    def get_max_depth(self):
        """ Gets the maximum expression depth.

        Returns:
            Max expression depth.
        """
        depth = 1
        stack = [[self, -1]]
        while stack:
            selem = stack[-1]
            expr = selem[0]
            if not isinstance(expr, CpoExpr):
                stack.pop()
            else:
                chlds = expr.children
                cnx = selem[1] + 1
                if cnx >= len(chlds):
                    depth = max(depth, len(stack))
                    stack.pop()
                else:
                    selem[1] = cnx
                    stack.append([chlds[cnx], -1])
        return depth


    def get_node_count(self):
        """ Gets the total expression node count.

        Node count is the total number of function calls and atoms in this expression.
        If a sub-expression is present multiple times, it is counted only once.

        Returns:
            Expression node count
        """
        return get_node_count(self)


    def pretty_print(self, out=None, mxdepth=None, mxargs=None, adsize=False, indent="", curdepth=0):
        """ Pretty print expression in a way that shows its structure.

        Args:
            out:      (optional) Print output (stream or file name, None for stdout)
            mxdepth:  (optional) Max print depth. Default is None (no limit)
            mxargs:   (optional) Max number of function or array arguments. Default is None (no limit)
            adsize:   (optional) Add size of expressions in  nodes
            indent:   (optional) Indentation string
            curdepth: (internal) Current print depth
        """
        pretty_print(self, out, mxdepth, mxargs, adsize, indent, curdepth)


    def __nonzero__(self):
        """ Safe function to protect for use of CpoExpr as boolean.

        This may occur when using comparison operators that create a CpoExpr instead of actually compare with
        another expression.

        Raises:
            CpoException as CpoExpr should not be used as boolean.
        """
        raise CpoException("CPO expression can not be used as boolean.")


    def __bool__(self):
        """ Safe function to protect for use of CpoExpr as boolean.

        This may occur when using comparison operators that create a CpoExpr instead of actually compare with
        another expression.

        Equivalent to __nonzero__ for Python 3

        Raises:
            CpoException as CpoExpr should not be used as boolean.
        """
        raise CpoException("CPO expression can not be used as boolean.")


    def equals(self, other):
        """ Checks the equality of this expression with another object.

        Implementation is required with a different name than __eq__ name because this function is already
        overloaded to construct model expression with operator '=='.

        Args:
            other: Other object to compare with.
        Return:
            True if 'other' is equal to this object, False otherwise.
        """
        return _is_equal_expressions(self, other)


    def _equals(self, other):
        """ Checks the equality of this expression with another object.

        This particular method just checks local attributes, but does not check recursively children if any.
        Recursion is implemented by method equals() that uses a self-managed stack to avoid too many
        recursive calls that may lead to an exception 'RuntimeError: maximum recursion depth exceeded'.

        Args:
            other: Other object to compare with.
        Return:
            True if 'other' is equal to this object, False otherwise.
        """
        return (type(self) == type(other)) and (self.type == other.type)


    def compare(self, other):
        """ Compare this expression with another.

        Args:
            other: Other CpoExpr to compare with.
        Returns:
            integer <0 if self < other, 0 if self == other, >0 if self > other
        """
        if self.type != other.type:
            return self.type.id - other.type.id
        if self.name != other.name:
            return self.type.id - other.type.id
        sclen = len(self.children)
        oclen = len(other.children)
        if sclen != oclen:
            return sclen - oclen
        for sc, oc in zip(self.children, other.children):
            r = sc.compare(oc)
            if r != 0:
                return r
        return 0


    def __str__(self):
        """ Convert this expression into a string """
        if is_string(self.name):
            return to_printable_string(self.name) + " = " + _to_string(self)
        else:
            return _to_string(self)

    # Operators overloading

    def __ne__(self, other):
        """ Not equal """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_FloatExpr) and other.is_kind_of(Type_FloatExpr), "Operands of != should be integer or float expressions"
        return CpoFunctionCall(Oper_diff, Type_BoolExpr, (self, other))

    def __eq__(self, other):
        """ Equal """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_FloatExpr) and other.is_kind_of(Type_FloatExpr), "Operands of == should be integer or float expressions"
        return CpoFunctionCall(Oper_equal, Type_BoolExpr, (self, other))

    def __gt__(self, other):
        """ Greater """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_FloatExpr) and other.is_kind_of(Type_FloatExpr), "Operands of > should be integer or float expressions"
        return CpoFunctionCall(Oper_greater, Type_BoolExpr, (self, other))

    def __ge__(self, other):
        """ Greater or equal """
        other = build_cpo_expr(other)
        if self.is_kind_of(Type_IntExpr):
            if other.is_kind_of(Type_FloatExpr):
                return CpoFunctionCall(Oper_greater_or_equal, Type_BoolExpr, (self, other))
            assert other.is_kind_of(Type_CumulExpr), "Operands of >= should be integer, float or cumul expressions"
            return CpoFunctionCall(Oper_greater_or_equal, Type_Constraint, (self, other))
        if self.is_kind_of(Type_FloatExpr):
            assert self.is_kind_of(Type_FloatExpr), "Operands of >= should be integer, float or cumul expressions"
            return CpoFunctionCall(Oper_greater_or_equal, Type_BoolExpr, (self, other))
        assert self.is_kind_of(Type_CumulExpr) and other.is_kind_of(Type_IntExpr), "Operands of >= should be integer, float or cumul expressions"
        return CpoFunctionCall(Oper_greater_or_equal, Type_Constraint, (self, other))

    def __lt__(self, other):
        """ Less """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_FloatExpr) and other.is_kind_of(Type_FloatExpr), "Operands of < should be integer or float expressions"
        return CpoFunctionCall(Oper_less, Type_BoolExpr, (self, other))

    def __le__(self, other):
        """ Less or equal """
        other = build_cpo_expr(other)
        if self.is_kind_of(Type_IntExpr):
            if other.is_kind_of(Type_FloatExpr):
                return CpoFunctionCall(Oper_less_or_equal, Type_BoolExpr, (self, other))
            assert other.is_kind_of(Type_CumulExpr), "Operands of <= should be integer, float or cumul expressions"
            return CpoFunctionCall(Oper_less_or_equal, Type_Constraint, (self, other))
        if self.is_kind_of(Type_FloatExpr):
            assert self.is_kind_of(Type_FloatExpr), "Operands of <= should be integer, float or cumul expressions"
            return CpoFunctionCall(Oper_less_or_equal, Type_BoolExpr, (self, other))
        assert self.is_kind_of(Type_CumulExpr) and other.is_kind_of(Type_IntExpr), "Operands of <= should be integer, float or cumul expressions"
        return CpoFunctionCall(Oper_less_or_equal, Type_Constraint, (self, other))

    def __add__(self, other):
        """ Plus """
        if other is 0:  # Do not use == because it is overloaded
            return self
        other = build_cpo_expr(other)
        if self.is_kind_of(Type_IntExpr):
            if other.is_kind_of(Type_IntExpr):
                return CpoFunctionCall(Oper_plus, Type_IntExpr, (self, other))
            assert other.is_kind_of(Type_FloatExpr), "Operands of + should be integer, float or cumul"
            return CpoFunctionCall(Oper_plus, Type_FloatExpr, (self, other))
        elif self.is_kind_of(Type_FloatExpr):
            assert other.is_kind_of(Type_FloatExpr), "Operands of + should be integer, float or cumul"
            return CpoFunctionCall(Oper_plus, Type_FloatExpr, (self, other))
        # Sum of cumul expressions
        assert self.is_kind_of(Type_CumulExpr) and other.is_kind_of(Type_CumulExpr), "Operands of + should be integer, float or cumul expressions"
        return CpoFunctionCall(Oper_plus, Type_CumulExpr, (self, other))

    def __radd__(self, other):
        """ Plus (right) """
        if other is 0:  # Do not use == because it is overloaded
            return self
        return build_cpo_expr(other).__add__(self)

    def __pos__(self):
        """ Unary plus """
        return self

    def __sub__(self, other):
        """ Minus """
        other = build_cpo_expr(other)
        if self.is_kind_of(Type_IntExpr):
            if other.is_kind_of(Type_IntExpr):
                return CpoFunctionCall(Oper_minus, Type_IntExpr, (self, other))
            assert other.is_kind_of(Type_FloatExpr), "Operands of - should be integer, float or cumul"
            return CpoFunctionCall(Oper_minus, Type_FloatExpr, (self, other))
        elif self.is_kind_of(Type_FloatExpr):
            assert other.is_kind_of(Type_FloatExpr), "Operands of - should be integer, float or cumul"
            return CpoFunctionCall(Oper_minus, Type_FloatExpr, (self, other))
        # Cumul expressions
        assert self.is_kind_of(Type_CumulExpr) and other.is_kind_of(Type_CumulExpr), "Operands of - should be integer, float or cumul"
        return CpoFunctionCall(Oper_minus, Type_CumulExpr, (self, other))

    def __rsub__(self, other):
        """ Minus (right) """
        return build_cpo_expr(other).__sub__(self)

    def __neg__(self):
        """ Unary minus """
        if self.is_kind_of(Type_IntExpr):
            return CpoFunctionCall(Oper_minus, Type_IntExpr, (self,))
        if self.is_kind_of(Type_FloatExpr):
            return CpoFunctionCall(Oper_minus, Type_FloatExpr, (self,))
        assert self.is_kind_of(Type_CumulExpr), "Operands of - should be integer, float or cumul"
        return CpoFunctionCall(Oper_minus, Type_CumulExpr, (self,))

    def __mul__(self, other):
        """ Multiply """
        # if other is 1: # Do not use == because it is overloaded
        #     return self
        other = build_cpo_expr(other)
        if self.is_kind_of(Type_IntExpr):
            if other.is_kind_of(Type_IntExpr):
                return CpoFunctionCall(Oper_times, Type_IntExpr, (self, other))
            assert other.is_kind_of(Type_FloatExpr), "Operands of * should be integer or float"
            return CpoFunctionCall(Oper_times, Type_FloatExpr, (self, other))
        assert other.is_kind_of(Type_FloatExpr), "Operands of * should be integer or float"
        return CpoFunctionCall(Oper_times, Type_FloatExpr, (self, other))

    def __rmul__(self, other):
        """ Multiply (right) """
        return build_cpo_expr(other).__mul__(self)

    def __div__(self, other):
        """ Float divide """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_FloatExpr) and other.is_kind_of(Type_FloatExpr), "Operands of / should be float or integer expressions"
        return CpoFunctionCall(Oper_float_div, Type_FloatExpr, (self, other))

    def __rdiv__(self, other):
        """ Float divide (right) """
        return build_cpo_expr(other).__div__(self)

    def __truediv__(self, other):
        """ Float divide (Python 3) """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_FloatExpr) and other.is_kind_of(Type_FloatExpr), "Operands of / should be float or integer expressions"
        return CpoFunctionCall(Oper_float_div, Type_FloatExpr, (self, other))

    def __rtruediv__(self, other):
        """ Float divide (right) (Python 3) """
        return build_cpo_expr(other).__truediv__(self)

    def __floordiv__(self, other):
        """ Integer divide (Python 3) """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_IntExpr) and other.is_kind_of(Type_IntExpr), "Operands of // should be integer expressions"
        return CpoFunctionCall(Oper_int_div, Type_IntExpr, (self, other))

    def __rfloordiv__(self, other):
        """ Integer divide (right) (Python 3) """
        return build_cpo_expr(other).__floordiv__(self)

    def __mod__(self, other):
        """ Modulo """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_IntExpr) and other.is_kind_of(Type_IntExpr), "Operands of % should be integer expressions"
        return CpoFunctionCall(Oper_mod, Type_IntExpr, (self, other))

    def __rmod__(self, other):
        """ Power (right) """
        return build_cpo_expr(other).__mod__(self)

    def __pow__(self, other):
        """ Power """
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_FloatExpr) and other.is_kind_of(Type_FloatExpr), "Operands of ** (power) should be float expressions"
        return CpoFunctionCall(Oper_power, Type_FloatExpr, (self, other))

    def __rpow__(self, other):
        """ Power (right) """
        return build_cpo_expr(other).__pow__(self)

    def __and__(self, other):
        """ Binary and used to represent logical and """
        # if other is True:  # Do not use == because it is overloaded
        #     return self
        # if other is False:  # Do not use == because it is overloaded
        #     return False
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_BoolExpr) and other.is_kind_of(Type_BoolExpr), "Operands of & (logical and) should be boolean expressions"
        return CpoFunctionCall(Oper_logical_and, Type_BoolExpr, (self, other))

    def __rand__(self, other):
        """ Binary and used to represent logical and (right) """
        return build_cpo_expr(other).__and__(self)

    def __or__(self, other):
        """ Binary or used to represent logical or """
        # if other is False:  # Do not use == because it is overloaded
        #     return self
        # if other is True:  # Do not use == because it is overloaded
        #     return True
        other = build_cpo_expr(other)
        assert self.is_kind_of(Type_BoolExpr) and other.is_kind_of(Type_BoolExpr), "Operands of | (logical or) should be boolean expressions"
        return CpoFunctionCall(Oper_logical_or, Type_BoolExpr, (self, other))

    def __ror__(self, other):
        """ Binary or used to represent logical or (right) """
        return build_cpo_expr(other).__or__(self)

    def __invert__(self):
        """ Binary not used to represent logical not (right) """
        assert self.is_kind_of(Type_BoolExpr), "Operands of ~ (logical not) should be a boolean expression"
        return CpoFunctionCall(Oper_logical_not, Type_BoolExpr, (self,))



class CpoValue(CpoExpr):
    """ CPO model expression representing a constant value. """
    __slots__ = ('value',  # Python value of the constant
                )

    def __init__(self, value, type):
        """ Constructor

        Args:
            value:  Constant value.
            type :  Value type.
        """
        assert isinstance(type, CpoType), "Argument 'type' should be a CpoType"
        super(CpoValue, self).__init__(type, None)
        if type.is_array_of_expr:
            self.children = value
        self.value = value

    def _equals(self, other):
        """ Checks the equality of this expression with another object.

        This particular method just checks local attributes, but does not check recursively children if any.
        Recursion is implemented by method equals() that uses a self-managed stack to avoid too many
        recursive calls that may lead to an exception 'RuntimeError: maximum recursion depth exceeded'.

        Args:
            other: Other object to compare with.
        Returns:
            True if 'other' is semantically identical to this object, False otherwise.
        """
        # Call super
        if not super(CpoValue, self)._equals(other):
            return False
        # For array of expr, already managed as a recursive call by _equals_expressions() called in super.
        if self.type.is_array_of_expr:
            return True
        # Check value
        return _is_equal_values(self.value, other.value)


class CpoAlias(CpoExpr):
    """ CPO model expression representing a symbolic alias on an expression. """
    __slots__ = ('expr',  # Target expression
                )

    def __init__(self, expr, name):
        """ Constructor

        Args:
            expr:  Target expression
            name:  Name of the alias
        """
        assert isinstance(expr, CpoExpr), "Argument 'expr' must be a CpoExpr"
        assert is_string(name), "Argument 'name' must be a string"
        super(CpoAlias, self).__init__(expr.type, name)
        self.expr = expr
        self.children = (expr,)


class CpoFunctionCall(CpoExpr):
    """ This class represent all model expression nodes that call a predefined modeler function.

    All modeling functions are available in module :mod:`docplex.cp.modeler`.
    """
    __slots__ = ('operation',  # Operation descriptor
                )

    def __init__(self, oper, rtype, oprnds):
        """ Constructor
        Args:
            oper:   Operation descriptor
            rtype:  Returned type
            oprnds: List of operand expressions.
        """
        assert isinstance(oper, CpoOperation), "Argument 'oper' should be a CpoOperation"
        super(CpoFunctionCall, self).__init__(rtype, None)
        self.operation = oper
        self.priority = oper.priority

        # Check no toplevel constraints
        for e in oprnds:
            assert not e.is_type(Type_Constraint), "A constraint can not be operand of an expression."
        self.children = oprnds

    def _equals(self, other):
        """ Checks the equality of this expression with another object.

        This particular method just checks local attributes, but does not check recursively children if any.
        Recursion is implemented by method equals() that uses a self-managed stack to avoid too many
        recursive calls that may lead to an exception 'RuntimeError: maximum recursion depth exceeded'.

        Args:
            other: Other object to compare with.
        Returns:
            True if 'other' is semantically identical to this object, False otherwise.
        """
        return super(CpoFunctionCall, self)._equals(other) and (self.operation == other.operation)


class CpoVariable(CpoExpr):
    """ This class is an abstract class extended by all expression nodes that represent a CPO variable.
    """
    __slots__ = ()

    def __init__(self, type, name):
        """ Constructor:

        Args:
            type:   Expression type.
            name:   Variable name.
        """
        super(CpoVariable, self).__init__(type, name)

    def is_variable(self):
        """ Check if this expression is a variable

        Returns:
            True if this expression is a variable
        """
        return True


class CpoIntVar(CpoVariable):
    """ This class represents an *integer variable* that can be used in a CPO model.

    This object should not be created explicitly, but using one of the following factory method:

    * :meth:`integer_var`, :meth:`integer_var_list`, :meth:`integer_var_dict` to create integer variable(s),
    * :meth:`binary_var`, :meth:`binary_var_list`, :meth:`binary_var_dict` to create integer variable(s)
      with value in [0..1],
    """
    __slots__ = ('domain',  # Variable domain
                 )

    def __init__(self, dom, name=None):
        # Private constructor
        super(CpoIntVar, self).__init__(Type_IntVar, name)
        self.domain = dom

    def set_domain(self, domain):
        """ Sets the domain of the variable.

        The domain of the variable is a list or tuple of:

           * discrete integer values,
           * list or tuple of 2 integers representing an interval.

        For example, here are valid domain definitions:

           set_domain([1, 3, 4, 5, 9])
           set_domain([1, (3, 5), 9])

        Args:
            domain: List of integers or interval tuples representing the variable domain.
        """
        self.domain = _build_int_var_domain(None, None, domain)

    def get_domain(self):
        """ Gets the domain of the variable.

        The domain of the variable can be:

         * A single integer value if the variable is fixed,
         * A list or a tuple of:

             - single integer values,
             - tuple of 2 integers representing an interval, bounds included

        Returns:
            Domain of the variable.
        """
        return self.domain

    def get_domain_min(self):
        """ Gets the domain lower bound.

        Returns:
            Domain lower bound.
        """
        return _domain_min(self.domain)

    def get_domain_max(self):
        """ Gets the domain upper bound.

        Returns:
            Domain upper bound.
        """
        return _domain_max(self.domain)

    def domain_iterator(self):
        """ Iterator on the individual values of an integer variable domain.

        Returns:
            Value iterator on the domain of this variable.
        """
        return _domain_iterator(self.domain)

    def domain_contains(self, value):
        """ Check whether a given value is in the domain of the variable

        Args:
            val: Value to check
        Returns:
            True if the value is in the domain, False otherwise
        """
        return _domain_contains(self.domain, value)

    @property
    def lb(self):
        """ This property is used to get lower bound of the variable domain.
        """
        return self.get_domain_min()

    @property
    def ub(self):
        """ This property is used to get upper bound of the variable domain.
        """
        return self.get_domain_max()

    def is_bool_var(self):
        """ Check if the domain of this variable is reduced to boolean, 0, 1

        Returns:
            True if this variable is a boolean variable
        """
        return self.domain == (0, 1) or self.domain == ((0, 1),)

    def is_binary(self):
        """ Check if the domain of this variable is reduced to boolean, 0, 1

        Returns:
            True if this variable is a boolean variable
        """
        return self.domain == (0, 1) or self.domain == ((0, 1),)

    def equals(self, other):
        """ Checks if this expression is equivalent to another

        Args:
            other: Other object to compare with.
        Return:
            True if 'other' is semantically identical to this object, False otherwise.
        """
        return super(CpoIntVar, self).equals(other) and (self.domain == other.domain)


class CpoBoolVar(CpoIntVar):
    # Currently internal
    # """ This class represents a *boolean variable* that can be used in a CPO model.
    #
    # This object should not be created explicitly, but using one of the following factory method:
    #
    # * :meth:`binary_var`, :meth:`binary_var_list`, :meth:`binary_var_dict` to create integer variable(s)
    #   with value in [0..1],
    # """
    __slots__ = ('domain',  # Variable domain
                 )

    def __init__(self, dom, name=None):
        # Private constructor
        super(CpoBoolVar, self).__init__(dom, name)
        self.type = Type_BoolVar

    def is_bool_var(self):
        """ Check if the domain of this variable is reduced to boolean, 0, 1

        Returns:
            True if this variable is a boolean variable
        """
        return True

    def is_binary(self):
        """ Check if the domain of this variable is reduced to boolean, 0, 1

        Returns:
            True if this variable is a boolean variable
        """
        return True


class CpoFloatVar(CpoVariable):
    # """ This class represents a *float variable* that can be used in a CPO model.
    # """
    __slots__ = ('min',  # Domain min value
                 'max',  # Domain max value
                 )

    def __init__(self, min, max, name=None):
        # Private constructor
        super(CpoFloatVar, self).__init__(Type_FloatVar, name)
        self.min = min
        self.max = max


    def get_domain_min(self):
        # """ Gets the domain lower bound.
        #
        # Returns:
        #     Domain lower bound.
        # """
        return self.min


    def get_domain_max(self):
        # """ Gets the domain upper bound.
        #
        # Returns:
        #     Domain upper bound.
        # """
        return self.max


    def get_domain(self):
        # """ Gets the variable domain as a tuple (min, max)
        #
        # Returns:
        #     Variable domain as a tuple (min, max)
        # """
        return (self.min, self.max,)


    def equals(self, other):
        # """ Checks if this expression is equivalent to another
        #
        # Args:
        #     other: Other object to compare with.
        # Return:
        #     True if 'other' is semantically identical to this object, False otherwise.
        # """
        return super(CpoFloatVar, self).equals(other) and (self.min == other.min) and (self.max == self.max)


###############################################################################
## Scheduling expressions
###############################################################################

INTERVAL_MAX = (INT_MAX // 2) - 1
""" Maximum interval variable range value """

INTERVAL_MIN = -INTERVAL_MAX
""" Minimum interval variable range value """

INFINITY = float('inf')
""" Infinity """

POSITIVE_INFINITY = float('inf')
""" Positive infinity """

NEGATIVE_INFINITY = float('-inf')
""" Negative infinity """

DEFAULT_INTERVAL = (0, INTERVAL_MAX)
""" Default interval. """

# Different interval variable presence states
_PRES_PRESENT   = "present"   # Always present
_PRES_ABSENT    = "absent"    # Always absent
_PRES_OPTIONAL  = "optional"  # Present or absent, choice made by the solver


class CpoIntervalVar(CpoVariable):
    """ This class represents an *interval variable* that can be used in a CPO model.

    This object should not be created explicitly, but using one of the following factory method
    :meth:`interval_var`, :meth:`interval_var_list`, or :meth:`interval_var_dict`.
    """
    __slots__ = ('start',        # Start domain
                 'end',          # End domain
                 'length',       # Length domain
                 'size',         # Size domain
                 'intensity',    # Specifies relation between size and length of the interval.
                 'granularity',  # Scale of the intensity function (int)
                 'presence',     # Presence requirement (in _PRES_*)
                 )

    def __init__(self, start, end, length, size, intensity, granularity, presence, name=None):
        # Private constructor
        super(CpoIntervalVar, self).__init__(Type_IntervalVar, name)
        self.start   = start
        self.end     = end
        self.length  = length
        self.size    = size
        self.presence = presence
        self.granularity = granularity
        self.intensity = intensity
        if intensity is not None:
            self.children = (intensity,)

    def set_start(self, intv):
        """ Sets the start interval.

        Args:
            intv: Start of the interval (single integer or interval expressed as a tuple of 2 integers).
        """
        self.start = _check_arg_interval(intv, "intv")

    def get_start(self):
        """ Gets the start of the interval.

        Returns:
            Start of the interval (interval expressed as a tuple of 2 integers).
        """
        return self.start

    def set_start_min(self, mn):
        """ Sets the minimum value of the start interval.

        Args:
            mn: Min value of the start of the interval.
        """
        assert _is_cpo_interval_value(mn), "Interval min should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mn <= self.start[1], "Interval min should be lower or equal to interval max"
        self.start = (mn, self.start[1])

    def set_start_max(self, mx):
        """ Sets the maximum value of the start interval.

        Args:
            mx: Max value of the start of the interval.
        """
        assert _is_cpo_interval_value(mx), "Interval max should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mx >= self.start[0], "Interval max should be lower or equal to interval max"
        self.start = (self.start[0], mx)

    def set_end(self, intv):
        """ Sets the end of the interval.

        Args:
            intv: End of the interval (single integer or interval expressed as a tuple of 2 integers).
        """
        self.end = _check_arg_interval(intv, "intv")

    def get_end(self):
        """ Gets the end of the interval.

        Returns:
            End of the interval (interval expressed as a tuple of 2 integers)
        """
        return self.end

    def set_end_min(self, mn):
        """ Sets the minimum value of the end interval.

        Args:
            mn: Min value of the end of the interval.
        """
        assert _is_cpo_interval_value(mn), "Interval min should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mn <= self.end[1], "Interval min should be lower or equal to interval max"
        self.end = (mn, self.end[1])

    def set_end_max(self, mx):
        """ Sets the maximum value of the end of the interval.

        Args:
            mx: Max value of the end of the interval.
        """
        assert _is_cpo_interval_value(mx), "Interval min should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mx >= self.end[0], "Interval max should be lower or equal to interval max"
        self.end = (self.end[0], mx)

    def set_length(self, intv):
        """ Sets the length interval.

        Args:
            intv: Length of the interval (single integer or interval expressed as a tuple of 2 integers).
        """
        self.length = _check_arg_interval(intv, "intv")

    def get_length(self):
        """ Gets the length interval.

        Returns:
            Length of the interval (interval expressed as a tuple of 2 integers).
        """
        return self.length

    def set_length_min(self, mn):
        """ Sets the minimum value of the length interval.

        Args:
            mn: Min value of the length of the interval min value.
        """
        assert _is_cpo_interval_value(mn), "Interval min should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mn <= self.length[1], "Interval min should be lower or equal to interval max"
        self.length = (mn, self.length[1])

    def set_length_max(self, mx):
        """ Sets the maximum value of the length interval.

        Args:
            mx: Max value of the length of the interval.
        """
        assert _is_cpo_interval_value(mx), "Interval max should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mx >= self.length[0], "Interval max should be lower or equal to interval max"
        self.length = (self.length[0], mx)

    def set_size(self, intv):
        """ Sets the size of the interval.

        Args:
            intv: Size of the interval (single integer or interval expressed as a tuple of 2 integers).
        """
        self.size = _check_arg_interval(intv, "intv")

    def get_size(self):
        """ Gets the size of the interval.

        Returns:
            Size of the interval (interval expressed as a tuple of 2 integers).
        """
        return self.size

    def set_size_min(self, mn):
        """ Sets the minimum value of the size interval.

        Args:
            mn: Min value of the size of the interval.
        """
        assert _is_cpo_interval_value(mn), "Interval min should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mn <= self.size[1], "Interval min should be lower or equal to interval max"
        self.size = (mn, self.size[1])

    def set_size_max(self, mx):
        """ Sets the maximum value of the size interval.

        Args:
            mx: Max value of the size of the interval.
        """
        assert _is_cpo_interval_value(mx), "Interval max should be an integer in [INTERVAL_MIN , INTERVAL_MAX]"
        assert mx >= self.size[0], "Interval max should be lower or equal to interval max"
        self.size = (self.size[0], mx)

    def set_present(self):
        """ Specifies that this IntervalVar must be present. """
        self.presence = _PRES_PRESENT

    def is_present(self):
        """ Check if this interval variable must be present.

        Returns:
            True if this interval variable must be present, False otherwise.
        """
        return self.presence == _PRES_PRESENT

    def set_absent(self):
        """ Specifies that this interval variable must be absent. """
        self.presence = _PRES_ABSENT

    def is_absent(self):
        """ Check if this interval variable must be absent.

        Returns:
            True if this interval variable must be absent, False otherwise.
        """
        return self.presence == _PRES_ABSENT

    def set_optional(self):
        """ Specifies that this interval variable is optional. """
        self.presence = _PRES_OPTIONAL

    def is_optional(self):
        """ Check if this interval variable is optional.

        Returns:
            True if this interval variable is optional, False otherwise.
        """
        return self.presence == _PRES_OPTIONAL

    def set_intensity(self, intensity):
        """ Sets the intensity function of this interval var.

        Args:
           intensity:  Intensity function (None, or StepFunction).
        """
        _check_arg_intensity(intensity, self.granularity)
        self.intensity = intensity
        if intensity is None:
            self.children = ()
        else:
            self.children = (intensity,)

    def get_intensity(self):
        """ Gets the intensity function of this interval var.

        Returns:
           Intensity function (None, or StepFunction).
        """
        return self.intensity

    def set_granularity(self, granularity):
        """ Sets the scale of the intensity function.

        Args:
            granularity: Scale of the intensity function (integer).
        """
        assert (granularity is None) or (is_int(granularity) and (granularity >= 0)), "Argument 'granularity' should be None or positive integer"
        self.granularity = granularity 

    def get_granularity(self):
        """ Get the scale of the intensity function.

        Returns:
            Scale of the intensity function, None for default (100)
        """
        return self.granularity

    def _equals(self, other):
        """ Checks the equality of this expression with another object.

        This particular method just checks local attributes, but does not check recursively children if any.
        Recursion is implemented by method equals() that uses a self-managed stack to avoid too many
        recursive calls that may lead to an exception 'RuntimeError: maximum recursion depth exceeded'.

        Args:
            other: Other object to compare with.
        Returns:
            True if 'other' is semantically identical to this object, False otherwise.
        """
        # Call super
        if not super(CpoIntervalVar, self)._equals(other):
            return False

        # Check same attributes (intensity processed as a children)
        return self.start == other.start and \
               self.end == other.end and \
               self.length == other.length and \
               self.size == other.size and \
               self.granularity == other.granularity and \
               self.presence == other.presence


class CpoSequenceVar(CpoVariable):
    """ This class represents an *sequence variable* that can be used in a CPO model.

    Variables are stored in 'children' attribute
    """
    __slots__ = ('types',  # Variable types
                )
    
    def __init__(self, vars, types=None, name=None):
        """ Creates a new sequence variable.

        This method creates an instance of sequence variable on the set of interval variables defined
        by the array 'vars'.
        A list of non-negative integer types can be optionally specified.
        List of variables and types must be of the same size and interval variable vars[i] will have type types[i]
        in the sequence variable.

        Args:
            vars:  List of IntervalVars that constitute the sequence.
            types: List of variable types as integers, same size as vars, or None (default).
            name:  Name of the sequence, None for automatic naming.
        """
        # Check  arguments
        if isinstance(vars, CpoValue):
            assert vars.is_kind_of(Type_IntervalVarArray)
            vars = vars.value
        else:
            assert is_array_of_type(vars, CpoIntervalVar), "Argument 'vars' should be an array of CpoIntervalVar"
        if types is not None:
            if isinstance(types, CpoValue):
                assert types.is_kind_of(Type_IntArray)
                types = types.value
            else:
                types = _check_and_expand_interval_tuples('types', types)
            assert len(types) == len(vars), "The array of types should have the same length than the array of variables."
        # Store attributes
        super(CpoSequenceVar, self).__init__(Type_SequenceVar, name)
        self.children = tuple(vars)
        self.types = types

    def get_interval_variables(self):
        """ Gets the array of variables.

        Returns:
            Array of interval variables that are in the sequence.
        """
        return self.children
    
    def get_vars(self):
        """ Gets the array of variables in this sequence variable.

        Returns:
            Array of interval variables
        """
        return self.children
    
    def get_types(self):
        """ Gets the array of types.

        Returns:
            Array of variable types (array of integers), None if no type defined.
        """
        return self.types

    def _equals(self, other):
        """ Checks the equality of this expression with another object.

        This particular method just checks local attributes, but does not check recursively children if any.
        Recursion is implemented by method equals() that uses a self-managed stack to avoid too many
        recursive calls that may lead to an exception 'RuntimeError: maximum recursion depth exceeded'.

        Args:
            other: Other object to compare with.
        Returns:
            True if 'other' is semantically identical to this object, False otherwise.
        """
        # Call super
        if not super(CpoSequenceVar, self)._equals(other):
            return False
        # Check equality of types
        if self.types != other.types:
            return False
        # List of variables is processed as expression children
        return True

    def __len__(self):
        """ Get the length of the sequence variable (number of variables) """
        return len(self.children)


class CpoTransitionMatrix(CpoValue):
    """ This class represents a *transition matrix* that is used in CPO model to represent transition distances.
    """
    __slots__ = ()  # Matrix stored in value field

    def __init__(self, size=None, values=None, name=None):
        """ Creates a new transition matrix (square matrix of integers).

        A transition matrix is a square matrix of non-negative integers that represents a minimal distance between
        two interval variables.
        An instance of transition matrix can be used in the no_overlap constraint and in state functions.

          * In a no_overlap constraint the transition matrix represents the minimal distance between two
            non-overlapping interval variables.
            The matrix is indexed using the integer types of interval variables in the sequence variable
            of the no_overlap constraint.

          * In a state function, the transition matrix represents the minimal distance between two integer
            states of the function.

        A transition matrix can be created:

          * Deprecated.
            Giving only its size. In this case, a transition matrix is created by this constructor with all
            values initialized to zero. Matrix values can then be set using :meth:`set_value` method.

          * Giving the matrix values as a list of rows, each row being a list of integers.
            Matrix values can not be changed after it has been created.

        Args:
            size (optional):   Matrix size (width and height),
            name (optional):   Name of the matrix. None by default.
            values (optional): Matrix values expressed as a list of rows.
        """

        super(CpoTransitionMatrix, self).__init__(None, Type_TransitionMatrix)
        if name:
            self.set_name(name)

        # Check type of argument
        if size:
            assert is_int(size) and size >= 0, "Argument 'size' should be a positive integer."
            assert values is None, "Arguments 'size' and 'values' should not be given together."
            warnings.warn("Creating editable transition matrix by size is deprecated since release 2.3.",
                          DeprecationWarning)
            self.value = [[0 for i in range(size)] for j in range(size)]
        else:
            try:
                self.value = tuple(tuple(x) for x in values)
            except TypeError:
                assert False, "Argument 'values' should be an iterable of iterables of integers."
            size = len(self.value)
            assert all (len(x) == size for x in self.value), \
                "Matrix value should be squared (list of rows of the same size)"
            assert all(all(is_int(v) and v >= 0 for v in r) for r in self.value), \
                "All matrix values should be positive integers"

    def get_size(self):
        """ Returns the size of the matrix.

        Returns:
            Matrix size.
        """
        return len(self.value)

    def get_value(self, from_state, to_state):
        """ Returns a value in the transition matrix.

        Args:
            from_state: Index of the from state.
            to_state:   Index of the to state.
        Returns:
            Transition value.
        """
        return self.value[from_state][to_state]

    def get_all_values(self):
        """ Returns an iterator on all matrix values, in row/column order

        Returns:
            Iterator on all values
        """
        sizerg = range(len(self.value))
        return (self.value[f][t] for f in sizerg for t in sizerg)

    def get_matrix(self):
        """ Returns the complete transition matrix.

        Returns:
            Transition matrix as a list of integers that is the concatenation of all matrix rows.
        """
        return self.value

    def set_value(self, from_state, to_state, value):
        """ Sets a value in the transition matrix.

        Args:
            from_state: Index of the from state.
            to_state:   Index of the to state.
            value:      Transition value.
        """
        assert is_int(value) and value >= 0, "Value should be a positive integer"
        self.value[from_state][to_state] = value


class CpoStateFunction(CpoVariable):
    """ This class represents a *state function* expression node.

    State functions are used by *interval variables* to represent the evolution of a state variable over time.
    """
    __slots__ = ('trmtx',  # Transition matrix
                )

    def __init__(self, trmtx=None, name=None):
        """ Creates a new state function.

        Args:
            trmtx (optional): An optional transition matrix defining the transition distance between consecutive states
                              of the state function.
                              Transition matrix is given as a list of rows (iterable of iterables of positive integers),
                              or as the result of a call to the method :meth:`~docplex.cp.expression.transition_matrix`.
            name (optional):   Name of the state function.
        """
        # Force name for state functions
        super(CpoStateFunction, self).__init__(Type_StateFunction, name)
        self.set_transition_matrix(trmtx)

    def set_transition_matrix(self, trmtx):
        """ Sets the transition matrix.

        Args:
        trmtx : A transition matrix defining the transition distance between consecutive states of the state function.
                Transition matrix is given as a list of rows (iterable of iterables of positive integers),
                or as the result of a call to the method :meth:`~docplex.cp.expression.transition_matrix`.
        """

        if trmtx is None:
            self.trmtx = None
            self.children = ()
        else:
            trmtx =  build_cpo_transition_matrix(trmtx)
            self.trmtx = trmtx
            assert isinstance(trmtx, CpoTransitionMatrix), "Argument 'trmtx' should be a CpoTransitionMatrix"
            self.children = (trmtx,)

    def get_transition_matrix(self):
        """ Returns the transition matrix.

        Returns:
            Transition matrix, None if none.
        """
        return self.trmtx

    def _equals(self, other):
        """ Checks the equality of this expression with another object.

        This particular method just checks local attributes, but does not check recursively children if any.
        Recursion is implemented by method equals() that uses a self-managed stack to avoid too many
        recursive calls that may lead to an exception 'RuntimeError: maximum recursion depth exceeded'.

        Args:
            other: Other object to compare with.
        Returns:
            True if 'other' is semantically identical to this object, False otherwise.
        """
        return super(CpoStateFunction, self)._equals(other)
        # Transition matrix is checked as children



###############################################################################
## Factory Functions
###############################################################################

def integer_var(min=None, max=None, name=None, domain=None):
    """ Creates an integer variable.

    An integer variable is a decision variable with a set of potential values called 'domain of the variable'.
    This domain can be expressed either:

     * as a single interval, with a minimum and a maximum bounds included in the domain,
     * or as an extensive list of values and/or intervals.

    When the domain is given extensively, an interval of the domain is represented by a tuple (min, max).
    Examples of variable domains expressed extensively are:

     * (1, 2, 3, 4)
     * (1, 2, (3, 7), 9)
     * ((1, 2), (7, 9))

    Following integer variable declarations are equivalent:

     * v = integer_var(0, 9, "X")
     * v = integer_var(domain=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9), name="X")
     * v = integer_var(domain=(0, (1, 5), (6, 7), 8, 9), name="X")
     * v = integer_var(domain=((0, 9)), name="X")

    Args:
        min:    Domain min value. Optional if domain is given extensively.
        max:    Domain max value. Optional if domain is given extensively.
        name:   Optional variable name. If not given, a name is automatically generated.
        domain: Variable domain expressed as extensive list of values and/or intervals expressed as tuples of integers.
                Unused if min and max are provided.
    Returns:
        CpoIntVar expression
    """
    return CpoIntVar(_build_int_var_domain(min, max, domain), name)


def integer_var_list(size, min=None, max=None, name=None, domain=None):
    """ Creates a list of integer variables.

    This methods creates a list of integer variables whose size is given as first parameter.
    All other parameters are identical to those requested by the method integer_var()
    that allows to create a single integer variable.
    See the documentation of :meth:`integer_var` for details.

    If a name is given, each variable of the list is created with this
    name concatenated with the index of the variable in the list, starting by zero.

    Args:
        size:   Size of the list of variables
        min:    Domain min value. Optional if domain is given extensively.
        max:    Domain max value. Optional if domain is given extensively.
        name:   Optional variable name prefix.
        domain: Variable domain expressed as extensive list of values and/or intervals expressed as tuples of integers.
                Unused if min and max are provided.
    Returns:
        List of integer variables.
    """
    dom = _build_int_var_domain(min, max, domain)
    res = []
    if name is None:
        for i in range(size):
            res.append(CpoIntVar(dom))
    else:
        name = name + "_"
        for i in range(size):
            res.append(CpoIntVar(dom, name + str(i)))
    return res


def integer_var_dict(keys, min=None, max=None, name=None, domain=None):
    """ Creates a dictionary of integer variables.

    This methods creates a dictionary of integer variables associated to a list of keys given as first parameter.
    All other parameters are identical to those requested by the method integer_var()
    that allows to create a single integer variable.
    See the documentation of :meth:`integer_var` for details.

    If a name is given, each variable of the list is created with this
    name concatenated with the string representation of the corresponding key.
    The parameter 'name' can also be a function that is called to build the variable name
    with the variable key as parameter.

    Args:
        keys:   Iterable of variable keys.
        min:    Domain min value. Optional if domain is given extensively.
        max:    Domain max value. Optional if domain is given extensively.
        name:   Optional variable name. If not given, a name is automatically generated.
        domain: Variable domain expressed as extensive list of values and/or intervals expressed as tuples of integers.
                Unused if min and max are provided.
    Returns:
        Dictionary of CpoIntVar objects.
    """
    dom = _build_int_var_domain(min, max, domain)

    res = {}
    if name is None:
        for k in keys:
            res[k] = CpoIntVar(dom)
    elif is_string(name):
        name = name + "_"
        for i, k in enumerate(keys):
            res[k] = CpoIntVar(dom, name + str(i))
    else:
        for k in keys:
            res[k] = CpoIntVar(dom, name=name(k))
    return res


def binary_var(name=None):
    """ Creates a binary integer variable.

    An binary variable is an integer variable with domain limited to 0 and 1

    Args:
        name (optional): Variable name, default is None for automatic name.
    Returns:
        CpoIntVar expression
    """
    return CpoIntVar(BINARY_DOMAIN, name)


def binary_var_list(size, name=None):
    """ Creates a list of binary variables.

    This methods creates a list of binary variables.

    If a name is given, each variable of the list is created with this
    name concatenated with the index of the variable in the list, starting by zero.

    Args:
        size: Size of the list of variables
        name (optional): Variable name prefix. If not given, a name prefix is generated automatically.
    Returns:
        List of binary integer variables.
    """
    return integer_var_list(size, BINARY_DOMAIN, name=name)


def binary_var_dict(keys, name=None):
    """ Creates a dictionary of binary variables.

    This methods creates a dictionary of binary variables associated to a list of keys given as first parameter.

    If a name is given, each variable of the list is created with this
    name concatenated with the string representation of the corresponding key.
    The parameter 'name' can also be a function that is called to build the variable name
    with the variable key as parameter.

    Args:
        keys: Iterable of variable keys.
        name (optional): Variable name prefix, or function to be called on dictionary key (example: str).
                         If not given, a name prefix is generated automatically.
    Returns:
        Dictionary of CpoIntVar objects (OrderedDict).
    """
    return integer_var_dict(keys, BINARY_DOMAIN, name=name)


def interval_var(start=None, end=None, length=None, size=None,
                 intensity=None, granularity=None, optional=False, name=None):
    """ Creates an interval variable.

    An interval decision variable represents an unknown of a scheduling problem, in particular an interval of time
    during which something happens (an activity is carried out) whose position in time is unknown.
    An interval is characterized by a start value, an end value and a size.
    The start and end of an interval variable must be in [INTERVAL_MIN..INTERVAL_MAX].
    An important feature of interval decision variables is that they can be optional, that is, it is possible
    to model that an interval variable can be absent from the solution schedule.

    Sometimes the intensity of work is not the same during the whole interval.
    For example, consider a worker who does not work during weekends (his work intensity during weekends is 0%)
    and on Friday he works only for half a day (his intensity during Friday is 50%).
    For this worker, 7 man-days work will span for longer than just 7 days.
    In this example 7 man-days represent what is called the size of the interval: that is, the length of the
    interval would be if the intensity function was always at 100%.
    To model such situations, a range for the size of an interval variable and an integer stepwise intensity
    function can be specified.
    The length of the interval will be at least long enough to cover the work requirements
    given by the interval size, taking into account the intensity function.

    Args:
        start (optional):       Allowed range for the start of the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        end (optional):         Allowed range for the end the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        length (optional):      Allowed range for the length the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        size (optional):        Allowed range for the size the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        intensity (optional):   StepFunction that specifies relation between size and length of the interval.
        granularity (optional): Scale of the intensity function.
        optional (optional):    Optional presence indicator.
        name (optional):        Name of the variable. If not given, a name is generated automatically.
    Returns:
        IntervalVar expression.
    """
    start  = _check_arg_interval(start,  "start")
    end    = _check_arg_interval(end,    "end")
    length = _check_arg_interval(length, "length")
    size   = _check_arg_interval(size,   "size")
    _check_arg_intensity(intensity, granularity)
    #presence = _PRES_OPTIONAL if (_check_arg_boolean(optional, "optional") or not present) else _PRES_PRESENT
    presence = _PRES_OPTIONAL if (_check_arg_boolean(optional, "optional")) else _PRES_PRESENT
    return CpoIntervalVar(start, end, length, size, intensity, granularity, presence, name)


def interval_var_list(asize, start=None, end=None, length=None, size=None,
                 intensity=None, granularity=None, optional=False, name=None):
    """ Creates a list of interval variables.

    If a name is given, each variable of the array is created with this
    name concatenated with the index of the variable in the list.

    Args:
        asize:                  Size of the list of variables
        start (optional):       Allowed range for the start of the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        end (optional):         Allowed range for the end the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        length (optional):      Allowed range for the length the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        size (optional):        Allowed range for the size the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        intensity (optional):   StepFunction that specifies relation between size and length of the interval.
        granularity (optional): Scale of the intensity function.
        optional (optional):    Optional presence indicator.
        name (optional):        Name of the variable. If not given, a name is generated automatically.
    Returns:
        List of interval variables.
    """
    start  = _check_arg_interval(start,  "start")
    end    = _check_arg_interval(end,    "end")
    length = _check_arg_interval(length, "length")
    size   = _check_arg_interval(size,   "size")
    _check_arg_intensity(intensity, granularity)
    presence = _PRES_OPTIONAL if (_check_arg_boolean(optional, "optional")) else _PRES_PRESENT

    res = []
    if name is None:
        for i in range(asize):
            res.append(CpoIntervalVar(start, end, length, size, intensity, granularity, presence))
    else:
        name = name + "_"
        for i in range(asize):
            res.append(CpoIntervalVar(start, end, length, size, intensity, granularity, presence, name + str(i)))
    return res


def interval_var_dict(keys, start=None, end=None, length=None, size=None,
                 intensity=None, granularity=None, optional=False, name=None):
    """ Creates a list of interval variables.

    If a name is given, each variable of the array is created with this
    name concatenated with the index of the variable in the list.

    Args:
        keys:                   Iterable of variable keys.
        start (optional):       Allowed range for the start of the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        end (optional):         Allowed range for the end the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        length (optional):      Allowed range for the length the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        size (optional):        Allowed range for the size the interval (single integer or interval expressed as a tuple of 2 integers).
                                Default value is [0..INTERVAL_MAX].
        intensity (optional):   StepFunction that specifies relation between size and length of the interval.
        granularity (optional): Scale of the intensity function.
        optional (optional):    Optional presence indicator.
        name (optional):        Variable name prefix, or function to be called on dictionary key (example: str).
                                If not given, a name prefix is generated automatically.
    Returns:
        Dictionary of CpoIntervalVar objects.
    """
    start  = _check_arg_interval(start,  "start")
    end    = _check_arg_interval(end,    "end")
    length = _check_arg_interval(length, "length")
    size   = _check_arg_interval(size,   "size")
    _check_arg_intensity(intensity, granularity)
    presence = _PRES_OPTIONAL if (_check_arg_boolean(optional, "optional")) else _PRES_PRESENT

    res = {}
    if name is None:
        for k in keys:
            res[k] = CpoIntervalVar(start, end, length, size, intensity, granularity, presence)
    elif is_string(name):
        name = name + "_"
        for i, k in enumerate(keys):
            res[k] = CpoIntervalVar(start, end, length, size, intensity, granularity, presence, name + str(i))
    else:
        for k in keys:
            res[k] = CpoIntervalVar(start, end, length, size, intensity, granularity, presence, name(k))
    return res


def sequence_var(vars, types=None, name=None):
    """ Creates a new sequence variable (list of interval variables).

    This method creates an instance of sequence variable on the set of interval variables defined
    by the array 'vars'.
    A list of non-negative integer types can be optionally  specified.
    List of variables and types must be of the same size and interval variable vars[i] will have type types[i]
    in the sequence variable.

    Args:
        vars:  List of IntervalVars that constitute the sequence.
        types: List of variable types as integers, same size as vars, or None (default).
        name:  Name of the sequence, None for automatic naming.
    Returns:
        IntervalVar expression.
    """
    return CpoSequenceVar(vars, types, name)


def float_var(min=NEGATIVE_INFINITY, max=POSITIVE_INFINITY, name=None):
    # """ Creates a float variable.
    #
    # Args:
    #     min:   (Optional) Domain min value. Default value is negative infinity.
    #     max:   (Optional) Domain max value. Default value is positive infinity.
    #     name:  (Optional) Name of the variable
    # Returns:
    #     CpoFloatVar expression
    # """
    # Check arguments
    assert is_number(min) and is_number(max), "Float var bounds should be numbers"
    assert min <= max, "Float var lower bound should be lower than upper bound"

    return CpoFloatVar(min, max, name)


def float_var_list(size, min=NEGATIVE_INFINITY, max=POSITIVE_INFINITY, name=None):
    # """ Creates a list of float variables.
    #
    # This methods creates a list of float variables whose size is given as first parameter.
    # All other parameters are identical to those requested by the method float_var()
    # that allows to create a single float variable.
    # See the documentation of :meth:`float_var` for details.
    #
    # If a name is given, each variable of the list is created with this
    # name concatenated with the index of the variable in the list, starting by zero.
    #
    # Args:
    #     size:   Size of the list of variables
    #     min:    (Optional) Domain min value. Default value is negative infinity.
    #     max:    (Optional) Domain max value. Default value is positive infinity.
    #     name:   (Optional) variable name prefix.
    # Returns:
    #     List of float variables (CpoFloatVar objects).
    # """
    # Check arguments
    assert is_number(min) and is_number(max), "Float var bounds should be numbers"
    assert min <= max, "Float var lower bound should be lower than upper bound"

    res = []
    if name is None:
        for i in range(size):
            res.append(CpoFloatVar(min, max))
    else:
        name = name + "_"
        for i in range(size):
            res.append(CpoFloatVar(min, max, name + str(i)))
    return res


def float_var_dict(keys, min=None, max=None, name=None):
    # """ Creates a dictionary of float variables.
    #
    # This methods creates a dictionary of float variables associated to a list of keys given as first parameter.
    # All other parameters are identical to those requested by the method float_var()
    # that allows to create a single float variable.
    # See the documentation of :meth:`float_var` for details.
    #
    # If a name is given, each variable of the list is created with this
    # name concatenated with the string representation of the corresponding key.
    # The parameter 'name' can also be a function that is called to build the variable name
    # with the variable key as parameter.
    #
    # Args:
    #     keys:   Iterable of variable keys.
    #     min:    Domain min value. Optional if domain is given extensively.
    #     max:    Domain max value. Optional if domain is given extensively.
    #     name:   Optional variable name. If not given, a name is automatically generated.
    # Returns:
    #     Dictionary of float variables (CpoFloatVar objects).
    # """
    # Check arguments
    assert is_number(min) and is_number(max), "Float var bounds should be numbers"
    assert min <= max, "Float var lower bound should be lower than upper bound"

    res = {}
    if name is None:
        for k in keys:
            res[k] = CpoFloatVar(min, max)
    elif is_string(name):
        name = name + "_"
        for i, k in enumerate(keys):
            res[k] = CpoFloatVar(min, max, name + str(i))
    else:
        for k in keys:
            res[k] = CpoFloatVar(min, max, name=name(k))
    return res


def transition_matrix(szvals, name=None):
    """ Creates a new transition matrix (square matrix of integers).

    A transition matrix is a square matrix of non-negative integers that represents a minimal distance between
    two interval variables.
    An instance of transition matrix can be used in the no_overlap constraint and in state functions.

      * In a no_overlap constraint the transition matrix represents the minimal distance between two
        non-overlapping interval variables.
        The matrix is indexed using the integer types of interval variables in the sequence variable
        of the no_overlap constraint.

      * In a state function, the transition matrix represents the minimal distance between two integer
        states of the function.

    A transition matrix can be created:

      * Deprecated.
        Giving only its size. In this case, a transition matrix is created by this constructor with all
        values initialized to zero. Matrix values can then be set using :meth:`set_value` method.

      * Giving the matrix values as a list of rows, each row being a list of integers.
        Matrix values can not be changed after it has been created.

    Args:
        szvals:  Matrix values expressed as a list of rows (iterable of iterables of positive integers).
        name (optional):  Name of the matrix. None by default.
    Returns:
        TransitionMatrix expression.
    """
    # Check build of editable matrix (deprecated)
    if is_int(szvals):
        return CpoTransitionMatrix(size=szvals, name=name)

    # Build matrix checking from cache
    res = build_cpo_transition_matrix(szvals)
    if name:
        res.set_name(name)
    return res


def tuple_set(tset, name=None):
    """ Create a tuple set.

    A tuple set is essentially a matrix of integers, but not necessarily square.
    Boolean expressions allowed_assignments() and forbidden_assignments() use a tuple set to express
    allowed/forbidden combinations of values for a collection of variables.
    Each allowed/forbidden combination is called a tuple and it is represented by a row in the tuple set matrix.

    Note that modeling methods allowed_assignments() and forbidden_assignments() automatically create a tuple_set
    object if given argument is an iterable of iterables.
    There is then no need to explicitly call this factory method to create a tuple set, except to assign a name to it.

    Args:
        tset:  List of tuples, as iterable of iterables of integers
        name:  Object name (default is None).
    Returns:
        TupleSet expression.
    """
    res = build_cpo_tupleset(tset)
    if name:
        res.set_name(name)
    return res


def state_function(trmtx=None, name=None):
    """ Create a new State Function

    Args:
        trmtx (optional): An optional transition matrix defining the transition distance between consecutive states
                          of the state function.
                          Transition matrix is given as a list of rows (iterable of iterables of positive integers),
                          or as the result of a call to the method :meth:`~docplex.cp.expression.transition_matrix`.
        name (optional):  Name of the state function
    Returns:
        CpoStateFunction expression
    """
    return CpoStateFunction(trmtx, name)


###############################################################################
##  Public Functions
###############################################################################

def is_cpo_expr(expr, type=None):
    """ Check if an expression is a CPO model expression

    Args:
        expr:             Value to check
        type (optional):  Precise CPO type, in Type_*
    Returns:
        True if parameter is a CPO expression, of the expected type if given
    """
    return isinstance(expr, CpoExpr) and (type is None or expr.type == type)


def get_node_count(expr):
    """ Get the number of model expression nodes in an expression or list of expressions

    Args:
        expr:   Expression or list of expressions
    Returns:
        Total number of nodes
    """
    # Initialize stack of expressions to parse
    estack = []
    try:
        estack.extend(expr)
    except:
        estack.append(expr)
    doneset = set()  # Set of ids of expressions already processed

    # Loop while expression stack is not empty
    while estack:
        e = estack.pop()
        eid = id(e)
        if not eid in doneset:
            doneset.add(eid)
            # Stack children expressions
            try:
               estack.extend(e.children)
            except:
                pass

    return len(doneset)


def pretty_print(expr, out=None, mxdepth=None, mxargs=None, adsize=False, indent="", curdepth=0):
    """ Pretty print expression in a way that shows its structure.

    Args:
        expr:     Expression to print
        out:      (optional) Print output (stream or file name, None for stdout)
        mxdepth:  (optional) Max print depth. Default is None (no limit)
        mxargs:   (optional) Max number of function or array arguments. Default is None (no limit)
        adsize:   (optional) Add size of expressions in nodes
        indent:   (optional) Indentation string
        curdepth: (internal) Current print depth
    """
    # Check file output
    if is_string(out):
        with open_utf8(os.path.abspath(out), mode='w') as f:
            pretty_print(expr, out, mxdepth, mxargs, adsize, indent, curdepth)
            return
    # Check default output
    if out is None:
        out = sys.stdout

    # Write start banner
    out.write(indent)

    # Check max depth reached
    if (mxdepth is not None) and (curdepth >= mxdepth):
        out.write("...")
        return

    # Check if tuple
    if isinstance(expr, tuple):
        _pretty_print_children('(', ')', expr, out, mxdepth, mxargs, adsize, indent, curdepth)
        return

    # Check list
    if isinstance(expr, list):
        _pretty_print_children('[', ']', expr, out, mxdepth, mxargs, adsize, indent, curdepth)
        return

    # Check if expression is not CpoExpr
    if not isinstance(expr, CpoExpr):
        out.write(str(expr))
        return

    # Check if expression has a name
    if expr.has_name():
        out.write(expr.get_name())
        return

    # Check expression type
    t = expr.type
    if t.is_constant:
        out.write(str(expr))
    elif t.is_variable:
        out.write(str(expr))
    elif t.is_array:
        _pretty_print_children('[', ']', expr.children, out, mxdepth, mxargs, adsize, indent, curdepth)
    else:
        _pretty_print_children(str(expr.operation.cpo_name) + '(', ')', expr.children, out, mxdepth, mxargs, adsize, indent, curdepth)


def _pretty_print_children(lstart, lend, lexpr, out, mxdepth, mxargs, adsize, indent, curdepth):
    """ Pretty print list of expression

    Args:
        lstart:   List start banner
        lend:     List end banner
        lexpr:    List of expression to take children from
        out:      Print output (stream or file name, None for stdout)
        mxdepth:  Max print depth. Default is None (no limit)
        mxargs:   Max number of function or array arguments. Default is None (no limit)
        adsize:   Add size of expressions in nodes
        indent:   Indentation string
        curdepth: Current print depth
    """
    # Write start banner
    out.write(lstart)
    nindent = indent + " | "
    if adsize:
        out.write(" size:" + str(get_node_count(lexpr)))

    # Write expressions
    nbargs = len(lexpr)
    curdepth += 1
    if (mxargs is None) or (nbargs <= mxargs):
        for x in lexpr:
            out.write('\n')
            pretty_print(x, out, mxdepth, mxargs, adsize, nindent, curdepth)
    else:
        eargs = mxargs // 2
        bargs = mxargs - eargs
        for x in lexpr[:bargs]:
            out.write('\n')
            pretty_print(x, out, mxdepth, mxargs, adsize, nindent, curdepth)
        out.write('\n')
        out.write(nindent)
        out.write("...(+" + str(nbargs - mxargs) + ")...")
        for x in lexpr[-eargs:]:
            out.write('\n')
            pretty_print(x, out, mxdepth, mxargs, adsize, nindent, curdepth)

    # Write end banner
    # out.write('\n')
    # out.write(indent)
    out.write(lend)


def _domain_min(d):
    """ Retrieves the lower bound of a domain

    Args:
        d: Domain
    Returns:
        Domain lower bound
    """
    if is_array(d):
        v = d[0]
        return v[0] if isinstance(v, tuple) else v
    return d


def _domain_max(d):
    """ Retrieves the upper bound of a domain

    Args:
        d: Domain
    Returns:
        Domain upper bound
    """
    if is_array(d):
        v = d[-1]
        return v[-1] if isinstance(v, tuple) else v
    return d


def _domain_iterator(d):
    """ Iterator on the individual values of an integer variable domain.

    Args:
        d: Domain to iterate
    Returns:
        Domain iterator
    """
    if isinstance(d, (list, tuple)):
        for x in d:
            if isinstance(x, (list, tuple)):
                min, max = x
                if min == max:
                    yield min
                else:
                    for v in range(min, max + 1):
                        yield v
            else:
                yield x
    else:
        yield d


def _domain_contains(d, val):
    """ Check whether a domain contains a given value

    Args:
        d:   Domain
        val: Value to check
    Returns:
        True if value is in the domain, False otherwise
    """
    if isinstance(d, (list, tuple)):
        for x in d:
            if isinstance(x, (list, tuple)):
                min, max = x
                if min <= val <= max:
                    return True
            elif x == val:
                return True
        return False
    return d == val


# Cache of CPO expressions corresponding to Python values
# This cache is used to retrieve the CPO expression that corresponds to a Python expression
# that is used multiple times in a model.
# This allows to:
#  - speed-up conversion as expression type has not to be recompute again
#  - reduce CPO file length as common expressions are easily identified.
_CACHE_CONTEXT = context.model.cache
_CPO_VALUES_FROM_PYTHON = ObjectCache(_CACHE_CONTEXT.size)
_CACHE_ACTIVE = _CACHE_CONTEXT.active

# Lock to protect the map
_CPO_VALUES_FROM_PYTHON_LOCK = threading.Lock()


class _CacheKeyTuple(tuple):
    """ Tuple that is used as a key of the expression cache """
    def __eq__(self, other):
        return isinstance(other, tuple) and len(other) == len(self) and all(x1 is x2 for x1, x2 in zip(self, other))
    def __hash__(self):
        return super(_CacheKeyTuple, self).__hash__()


def _convert_to_tuple_if_possible(val):
    try:
        return tuple(val)
    except TypeError:
        return val


def build_cpo_expr(val):
    """ Builds an expression from a given Python value.

    This method uses a cache to return the same CpoExpr for the same constant.

    Args:
        val: Value to convert (possibly already an expression).
    Returns:
        Corresponding expression.
    Raises:
        CpoException if conversion is not possible.
    """
    # Check if already a CPO expression
    vtyp = type(val)
    if issubclass(vtyp, CpoExpr):
        return val

    #  Check atoms (not cached)
    ctyp = _PYTHON_TO_CPO_TYPE.get(vtyp)
    if ctyp:
        return CpoValue(val, ctyp)

    # Check numpy scalars (special case when called from overloaded operator)
    if vtyp is NUMPY_NDARRAY and not val.shape:
        return CpoValue(val, _PYTHON_TO_CPO_TYPE.get(val.dtype.type))

    # Value is any type of array. Force it as a tuple
    try:
        val = _CacheKeyTuple(val)
    except TypeError:
       raise CpoException("Impossible to build a CP Optimizer expression with value '{}' of type '{}'".format(to_string(val), type(val)))

    # Check if already in the cache
    if _CACHE_ACTIVE:
        with _CPO_VALUES_FROM_PYTHON_LOCK:
            try:
                cpval = _CPO_VALUES_FROM_PYTHON.get(val)
            except TypeError:
                # Convert to tuple every member of the tuple
                val = _CacheKeyTuple(_convert_to_tuple_if_possible(x) for x in val)
                try:
                    cpval = _CPO_VALUES_FROM_PYTHON.get(val)
                except TypeError:
                    raise CpoException("Impossible to build a CP Optimizer expression with value '{}' of type '{}'".format(to_string(val), type(val)))
            if cpval is None:
                cpval = _create_cpo_array_expr(val)
                _CPO_VALUES_FROM_PYTHON.set(val, cpval)

    else:
        cpval = _create_cpo_array_expr(val)

    return cpval


def build_cpo_tupleset(val):
    """ Builds a TupleSet expression from a Python value.

    This method uses the value cache to return the same CpoExpr for the same value.

    Args:
        val: Value to convert. Iterator or iterators of integers, or existing TupleSet expression.
    Returns:
        Model tupleset, not editable.
    Raises:
        Exception if conversion is not possible.
    """
    # Check if already a TupleSet expression
    if isinstance(val, CpoExpr) and val.is_type(Type_TupleSet):
        return val

    # Create result set
    try:
        tset = tuple(tuple(x) for x in val)
    except TypeError:
        assert False, "Argument should be an iterable of iterables of integers."

    # Check if already in the cache
    if _CACHE_ACTIVE:
        with _CPO_VALUES_FROM_PYTHON_LOCK:
            key = ('tupleset', tset)
            cpval = _CPO_VALUES_FROM_PYTHON.get(key)
            if cpval is None:
                # Verify new tuple set
                if tset:
                    #assert len(tset) > 0, "Tuple set should not be empty"
                    size = len(tset[0])
                    assert all(len(t) == size for t in tset), "All tuples in 'tset' should have the same length"
                    assert all(all(is_int(v) for v in r) for r in tset), "All tupleset values should be integer"
                # Put tuple set in cache
                cpval = CpoValue(tset, Type_TupleSet)
                _CPO_VALUES_FROM_PYTHON.set(key, cpval)
    else:
        cpval = CpoValue(tset, Type_TupleSet)

    return cpval


def build_cpo_transition_matrix(val):
    """ Builds a TransitionMatrix expression from a Python value.

    This method uses the value cache to return the same CpoExpr for the same value.

    Args:
        val: Value to convert. Iterator or iterators of integers, or existing TransitionMatrix expression.
    Returns:
        Model transition matrix, not editable.
    Raises:
        Exception if conversion is not possible.
    """
    # Check if already a TransitionMatrix expression
    if isinstance(val, CpoExpr) and val.is_type(Type_TransitionMatrix):
        return val

    # Create internal tuple
    try:
        trmx = tuple(tuple(x) for x in val)
    except TypeError:
        assert False, "Argument should be an iterable of iterables of integers."

    # Check if already in the cache
    if _CACHE_ACTIVE:
        with _CPO_VALUES_FROM_PYTHON_LOCK:
            key = ('matrix', trmx)
            cpval = _CPO_VALUES_FROM_PYTHON.get(key)
            if cpval is None:
                # Verify matrix
                size = len(trmx)
                assert size > 0, "Transition matrix should not be empty"
                assert all(len(t) == size for t in trmx), "All matrix lines should have the same length " + str(size)
                assert all(all(is_int(v) and v >= 0 for v in r) for r in trmx), "All matrix values should be positive integer"
                # Build matrix and put it in cache
                cpval = CpoTransitionMatrix(values=trmx)
                _CPO_VALUES_FROM_PYTHON.set(key, cpval)
    else:
        cpval = CpoTransitionMatrix(values=trmx)

    return cpval


_VAR_TYPE_ORDER = (Type_Bool, Type_Int, Type_Float, Type_IntArray, Type_FloatArray, Type_TupleSet,
                   Type_SegmentedFunction, Type_StepFunction, Type_TransitionMatrix,
                   Type_IntVar, Type_FloatVar, Type_StateFunction, Type_IntervalVar, Type_SequenceVar)
def compare_expressions(x1, x2):
    """ Compare two expressions for declaration order

    Args:
        x1: Expression 1
        x2: Expression 2
    Returns:
        Integer value that is negative if v1 < v2, zero if v1 == v2 and positive if v1 > v2.
    """
    # First sort by expression type
    if x1.type != x2.type:
        return _VAR_TYPE_ORDER.index(x1.type) - _VAR_TYPE_ORDER.index(x2.type)
    # Check object type
    tx1 = type(x1)
    tx2 = type(x2)
    if tx1 is not tx2:
        # Alias always lost
        if tx1 is CpoAlias:
            return 1
        if tx2 is CpoAlias:
            return -1
    # Compare by name in natural order
    return compare_natural(x1.get_name(), x2.get_name())


###############################################################################
## Private utility functions
###############################################################################

def _create_cpo_array_expr(val):
    """ Create a new CP expression from a given array Python value

    Args:
        val: Origin value, as a _CacheKeyTuple(val)
    Returns:
        New expression
    Raises:
        CpoException if it is not possible.
    """
    # Determine type
    typ = _get_cpo_array_type(val)
    if typ is None:
        raise CpoException("Impossible to build a CP Optimizer expression with value '" + to_string(val) + "'")

    # Convert array elements if required
    if typ.is_array_of_expr:
        return CpoValue(tuple(build_cpo_expr(v) for v in val), typ)

    # Default
    return CpoValue(val, typ)


def _get_cpo_array_type(val):
    """ Determine the CPO type of a given Python array value
    Args:
        val: Python value
    Returns:
        Corresponding CPO Type, None if none
    """
    # Check empty Array
    if len(val) == 0:
        return Type_IntArray

    # Get the most common type for all array elements
    cet = None
    for v in val:
        # Determine type of element
        nt = _get_cpo_type(v)
        if nt is None:
            return None
        # Combine with global type
        ncet = nt if cet is None else cet.get_common_type(nt)
        if ncet is None:
            # Check special case of couple of int mixed with ints
            if cet is Type_Int:
                if (nt is Type_IntArray) and _is_cpo_int_interval(v):
                    ncet = Type_Int
                else:
                    return None
            elif cet is Type_IntArray:
                if nt is Type_Int:
                    ncet = Type_Int
                else:
                    return None
            else:
               return None
        cet = ncet

    # Determine array type for result element type
    if cet is Type_IntArray:
        if all(_is_cpo_int_interval(v) for v in val):
            return Type_IntArray
        else:
            return Type_TupleSet
    return cet.parent_array_type


def _get_cpo_type(val):
    """ Determine the CPO type of a given Python value
    Args:
        val: Python value
    Returns:
        Corresponding CPO Type, None if none
    """
    # Check simple types
    ctyp = _PYTHON_TO_CPO_TYPE.get(type(val))
    if ctyp:
        return ctyp

    # Check CPO Expr
    if isinstance(val, CpoExpr):
        return val.type

    # Check numpy Array Scalars (special case when called from overloaded operator)
    if type(val) is NUMPY_NDARRAY and not val.shape:
        return _PYTHON_TO_CPO_TYPE.get(val.dtype.type)

    # Check array
    if is_array(val):
        return _get_cpo_array_type(val)

    return None


def _clear_value_cache():
    """ Clear the cache of CPO values
    """
    with _CPO_VALUES_FROM_PYTHON_LOCK:
        _CPO_VALUES_FROM_PYTHON.clear()


def _get_cpo_type_str(val):
    """ Get the CPO type name of a value

    Args:
        val: Value
    Returns:
        Value type string in CPO types
    """
    return _get_cpo_type(val).get_name()


def _create_operation(oper, params):
    """ Create a new expression that matches an operation descriptor

    Search in the signatures which one matches a set or arguments
    and then create an instance of the returned expression

    Args:
        oper:   Operation descriptor
        params: List of expression parameters
    Returns:
        New expression 
    Raises:
        CpoException if no operation signature matches arguments
    """
    # assert isinstance(oper, CpoOperation)

    # Convert arguments in CPO expressions
    args = tuple(map(build_cpo_expr, params))
      
    # Search corresponding signature
    s = _get_matching_signature(oper, args)
    if s is None:
        raise CpoException("The combination of parameters ({}) is not allowed for operation '{}' ({})"
                           .format(", ".join(map(_get_cpo_type_str, args)), oper.python_name, oper.cpo_name))

    # Create result expression
    return CpoFunctionCall(s.operation, s.return_type, args)


###############################################################################
##  Private Functions
###############################################################################

# Mapping of Python types to CPO types
_PYTHON_TO_CPO_TYPE = {}
for t in BOOL_TYPES:
    _PYTHON_TO_CPO_TYPE[t] = Type_Bool
for t in INTEGER_TYPES:
    _PYTHON_TO_CPO_TYPE[t] = Type_Int
for t in FLOAT_TYPES:
    _PYTHON_TO_CPO_TYPE[t] = Type_Float
_PYTHON_TO_CPO_TYPE[CpoIntVar]           = Type_IntVar
_PYTHON_TO_CPO_TYPE[CpoIntervalVar]      = Type_IntervalVar
_PYTHON_TO_CPO_TYPE[CpoSequenceVar]      = Type_SequenceVar
_PYTHON_TO_CPO_TYPE[CpoTransitionMatrix] = Type_TransitionMatrix
_PYTHON_TO_CPO_TYPE[CpoStateFunction]    = Type_StateFunction


def _is_cpo_int(arg):
    """ Check that a value is a valid integer for cpo
    Args:
        arg:  Value to check
    Returns:
        True if value is a valid CPO integer, False otherwise
    """
    return is_int(arg) and (INT_MIN <= arg <= INT_MAX)


def _is_cpo_int_interval(val):
    """ Check if a value represents an interval of CPO integers
    Args:
        val:  Value to check
    Returns:
        True if value is a tuple representing an interval of integers
    """
    return isinstance(val, tuple) and (len(val) == 2) and _is_cpo_int(val[0]) and _is_cpo_int(val[1]) and (val[1] >= val[0])


def _is_cpo_interval_value(arg):
    """ Check that a value is a valid value for an interval
    Args:
        arg:  Value to check
    Returns:
        True if value is a valid interval, False otherwise
    """
    return is_int(arg) and (INTERVAL_MIN <= arg <= INTERVAL_MAX)


def _check_arg_boolean(arg, name):
    """ Check that an argument is a boolean and raise error if wrong
    Args:
        arg:  Argument value
        name: Argument name
    Returns:
        Boolean to be set
    Raises:
        Exception if argument has the wrong format
    """
    assert is_bool(arg), "Argument '{}' should be a boolean".format(name)
    return arg


def _check_arg_integer(arg, name):
    """ Check that an argument is an valid integer value and raises error if wrong
    Args:
        arg:  Argument value
        name: Argument name
    Returns:
        Integer to be set
    Raises:
        Exception if argument has the wrong format
    """
    assert is_int(arg) and (INT_MIN <= arg <= INT_MAX), "Argument '{}' should be an integer in [INT_MIN , INT_MAX]".format(name)
    return arg


def _check_arg_interval(arg, name):
    """ Check that an argument is an interval and raise error if wrong
    Args:
        arg:  Argument value
        name: Argument name
    Returns:
        Interval to be set
    Raises:
        Exception if argument has the wrong format
    """
    # Check default interval
    if arg is None:
        return DEFAULT_INTERVAL

    # Check single value
    if is_int(arg):
        assert INTERVAL_MIN <= arg <= INTERVAL_MAX, "Argument '{}' should be in [INTERVAL_MIN , INTERVAL_MAX]".format(name)
        return arg, arg

    assert isinstance(arg, (list, tuple)) and len(arg) == 2, "Argument '" + name + "' should be an integer or an interval expressed as a tuple of two integers"
    lb, ub = arg
    assert _is_cpo_interval_value(lb), "Lower bound of argument '{}' should be an integer in [INTERVAL_MIN , INTERVAL_MAX]".format(name)
    assert _is_cpo_interval_value(ub), "Upper bound of argument '{}' should be an integer in [INTERVAL_MIN , INTERVAL_MAX]".format(name)
    assert lb <= ub, "Lower bound or argument '{}' should be lower or equal to upper bound".format(name)
    return arg


def _build_int_var_domain(min, max, domain):
    """ Create/check integer variable domain from integer variable creation parameters
    Args:
        min:    Domain min value, None if extensive list.
        max:    Domain max value, None if extensive list.
        domain: Extensive list of values and/or intervals expressed as tuples of integers.
    Returns:
        Valid integer variable domain
    Raises:
        Exception if argument has the wrong format
    """
    # Domain not given extensively
    if domain is None:
        if min is None:
            if max is None:
                return ((INT_MIN, INT_MAX),)
            _check_arg_integer(max, "max")
            return ((INT_MIN, max),)
        else:
            if max is None:
                # Test that first argument is directly a domain (ascending compatibility)
                if is_array(min):
                    domain = min
                    min = None
                else:
                    _check_arg_integer(min, "min")
                    return ((min, INT_MAX),)
            else:
                _check_arg_integer(min, "min")
                _check_arg_integer(max, "max")
                assert min <= max, "Argument 'min' should be lower or equal to 'max'"
                return ((min, max),)

    # Domain given extensively
    assert (min is None) and (max is None), "If domain is given extensively in 'domain', 'min' and/or 'max' should not be given"
    if is_int(domain):
        return domain
    domain = tuple(domain)
    assert all(_is_cpo_int(v) or _is_cpo_int_interval(v) for v in domain), "Argument 'domain' should be a list of integers and/or intervals (tuples of 2 integers)"
    return domain


def _check_arg_step_function(arg, name):
    """ Check that an argument is a step function and raise error if wrong
    Args:
        arg:  Argument value
        name: Argument name
    Returns:
        Resulting step function
    Raises:
        Exception if argument has the wrong format
    """
    assert isinstance(arg, CpoExpr) and (arg.type == Type_StepFunction), "Argument '" + name + "' should be a CpoStepFunction"
    return arg


def _check_arg_intensity(intensity, granularity):
    """ Check the intensity parameter of an interval var.
    Args:
       intensity:   Intensity function (None, or StepFunction).
       granularity: Granularity
    """
    if __debug__ and (intensity is not None):
        assert isinstance(intensity, CpoExpr) and (intensity.is_type(Type_StepFunction)), "Interval variable 'intensity' should be None or a CpoStepFunction"
        if granularity is None:
            granularity = 100
        for (s, v) in intensity.get_step_list():
            assert is_int(s), "'intensity' step start should be an integer"
            assert is_int(v) and (v >= 0) and (v <= granularity), "'intensity' step value should be in [0..granularity]"


def _check_and_expand_interval_tuples(name, arr):
    """ Check that a list contains only integers and expand interval tuples if any
    Args:
        name:  Argument name
        arr:   Array of integers and/or intervals
    Returns:
        Array of integers
    Raises:
        Exception if wrong type
    """
    assert isinstance(arr, (list, tuple)), "Argument '{}' (type {}) should be a list of integers or intervals".format(name, type(arr))
    res = None
    for i in range(len(arr)):
        v = arr[i]
        if is_int(v):
            if res:
                res.append(v)
        else:
            assert _is_cpo_int_interval(v), "Argument '{}' (type {}) should be a list of integers or intervals".format(name, type(arr))
            if not res:
                res = list(arr[:i])
            res.extend(range(v[0], v[1] + 1))
    return tuple(res if res else arr)


def _get_matching_signature(oper, args):
    """ Search the first operation signature matched by a list of arguments

    Args:
        oper: Operation where searching signature
        args: Candidate list of argument expressions
    Returns:
        Matching signature, None if not found
    """
    # Search corresponding signature
    return next((s for s in oper.signatures if _is_matching_arguments(s, args)), None)


def _is_matching_arguments(sgn, args):
    """ Check if a list of argument expressions matches this signature

    Args:
        sgn:  Signature descriptor
        args: Candidate list of argument expressions
    Returns:
        True if the arguments are matching signature
    """
    params = sgn.parameters
    if params is ANY_ARGUMENTS:
        return True
    for a, p in zip_longest(args, params):
        if a is not None:
            # Accepted if there is a parameter descriptor that is compatible with argument type
            if not (p and a.type.is_kind_of(p.type)):
                return False
        else:
            # Argument absent, check that parameter has a default value
            if p.default_value is None:
                return False
    return True


def _is_equal_expressions(v1, v2):
    """ Check if two expressions can be considered as equivalent.

    This method handles values that can be CPO expressions, None, and manage possible
    differences between number representations.

    It is implemented outside expression objects and use a self-managed stack to avoid too many
    recursive calls that may lead to an exception 'RuntimeError: maximum recursion depth exceeded'.

    Args:
        v1:  First CPO expression
        v2:  Second CPO expression
    Returns:
        True if both values are 'equivalent'
    """
    # Initialize expression stack
    estack = [[v1, v2, -1]]  # [expr1, expr2, child index]

    # Loop while expression stack is not empty
    while estack:
        # Get expressions to compare
        edscr = estack[-1]
        v1, v2, cx = edscr
        #print("Compare {} (type {}) with {} (type {})".format(v1, type(v1), v2, type(v2)))

        # Skip aliases
        while isinstance(v1, CpoAlias):
            v1 = v1.expr
        while isinstance(v2, CpoAlias):
            v2 = v2.expr

        # Check physical equality
        if v1 is v2:
            estack.pop()
            continue

        # Check same object type
        if type(v1) != type(v2):
            return False

        # Check if expression is a CPO expression
        if isinstance(v1, CpoExpr):
            # Check objects itself
            if (cx < 0) and not v1._equals(v2):
                return False
            # Access children
            ar1 = v1.children
            ar2 = v2.children
            # Check children
            alen = len(ar1)
            if (cx < 0) and (alen != len(ar2)):
                return False
            cx += 1
            if cx >= alen:
                estack.pop()
                continue
            # Store new child index in descriptor
            edscr[2] = cx
            # Stack children to compare it
            estack.append([ar1[cx], ar2[cx], -1])

        # Else, expressions are Python values
        else:
            if not _is_equal_values(v1, v2):
                return False
            estack.pop()

    # Expressions identical
    return True


def _is_equal_values(v1, v2):
    """ Check if two values can be considered as equivalent.

    This method handles values that can be CPO expressions, None, and manage possible
    differences between number representations.

    Args:
        v1:  First value
        v2:  Second value
    Returns:
        True if both values are 'equivalent'
    """
    # Check obvious cases
    if v1 is v2:
        return True
    # Check specifically CPO expressions (to not call '==' operator on it)
    if isinstance(v1, CpoExpr):
        return v1.equals(v2)
    if isinstance(v2, CpoExpr):
        return False
    # Check floats
    if is_float(v1):
        return is_float(v2) and (abs(v1 - v2) <= _FLOATING_POINT_PRECISION * max(abs(v1), abs(v2)))
    if is_array(v1):
        return is_array(v2) and (len(v1) == len(v2)) and all(_is_equal_values(x1, x2) for x1, x2 in zip(v1, v2))
    if isinstance(v1, dict):
        return isinstance(v2, dict) and (len(v1) == len(v2)) and all(_is_equal_values(v1[k], v2[k]) for k in v1)
    # Check finally basic equality
    return v1 == v2


def _reset_cache():
    """ Reset expression cache (for testing purpose).
    """
    _CPO_VALUES_FROM_PYTHON.clear()


def _to_string(expr):
    """ Build a string representing an expression.
    Args:
        expr:  Expression to convert into string
    Returns:
        String representing this expression
    """
    import docplex.cp.cpo.cpo_compiler as compiler
    return compiler.expr_to_string(expr)
